"""Lookup tools

These classes enable efficient extraction of precomputed lookup tables
from multiple source file formats into a uniform function-call accessor.
"""

from .extractor import extractor
from .evaluator import evaluator

__all__ = [
    "extractor",
    "evaluator",
]



import awkward as ak
from coffea import hist, processor
from coffea import nanoevents


class NanoEventsProcessor(processor.ProcessorABC):
    def __init__(self, columns=[], canaries=[]):
        self._columns = columns
        self._canaries = canaries
        dataset_axis = hist.Cat("dataset", "Primary dataset")
        mass_axis = hist.Bin("mass", r"$m_{\mu\mu}$ [GeV]", 30000, 0.25, 300)
        pt_axis = hist.Bin("pt", r"$p_{T}$ [GeV]", 30000, 0.25, 300)

        self.expected_usermeta = {
            "ZJets": ("someusermeta", "hello"),
            "Data": ("someusermeta2", "world"),
        }

        self._accumulator = processor.dict_accumulator(
            {
                "mass": hist.Hist("Counts", dataset_axis, mass_axis),
                "pt": hist.Hist("Counts", dataset_axis, pt_axis),
                "cutflow": processor.defaultdict_accumulator(int),
                "worker": processor.set_accumulator(),
            }
        )

    @property
    def columns(self):
        return self._columns

    @property
    def accumulator(self):
        return self._accumulator

    def process(self, events):
        output = self.accumulator.identity()

        dataset = events.metadata["dataset"]
        print(events.metadata)
        if "checkusermeta" in events.metadata:
            metaname, metavalue = self.expected_usermeta[dataset]
            assert metavalue == events.metadata[metaname]

        mapping = events.behavior["__events_factory__"]._mapping
        muon_pt = events.Muon.pt
        if isinstance(mapping, nanoevents.mapping.CachedMapping):
            keys_in_cache = list(mapping.cache.cache.keys())
            has_canaries = [canary in keys_in_cache for canary in self._canaries]
            if has_canaries:
                try:
                    from distributed import get_worker

                    worker = get_worker()
                    output["worker"].add(worker.name)
                except ValueError:
                    pass

        dimuon = ak.combinations(events.Muon, 2)
        dimuon = dimuon["0"] + dimuon["1"]

        output["pt"].fill(dataset=dataset, pt=ak.flatten(muon_pt))
        output["mass"].fill(dataset=dataset, mass=ak.flatten(dimuon.mass))
        output["cutflow"]["%s_pt" % dataset] += sum(ak.num(events.Muon))
        output["cutflow"]["%s_mass" % dataset] += sum(ak.num(dimuon))

        return output

    def postprocess(self, accumulator):
        return accumulator



from .base import BaseSchema
from .nanoaod import NanoAODSchema, PFNanoAODSchema
from .treemaker import TreeMakerSchema
from .physlite import PHYSLITESchema
from .delphes import DelphesSchema
from .pdune import PDUNESchema

__all__ = [
    "BaseSchema",
    "NanoAODSchema",
    "PFNanoAODSchema",
    "TreeMakerSchema",
    "PHYSLITESchema",
    "DelphesSchema",
    "PDUNESchema",
]



"""Basic NanoEvents and NanoCollection mixins"""
import awkward
import numpy
from abc import abstractmethod
from functools import partial
import re
import coffea
from coffea.util import rewrap_recordarray, awkward_rewrap
from typing import (
    List,
    Union,
    Tuple,
    Callable,
)


behavior = {}


@awkward.mixin_class(behavior)
class Systematic:
    """A base mixin class to describe and build variations on a feature of an nanoevents object."""

    _systematic_kinds = set()

    @classmethod
    def add_kind(cls, kind: str):
        """
        Register a type of systematic variation, it must fullfil the base class interface.
        """
        cls._systematic_kinds.add(kind)

    def _ensure_systematics(self):
        """
        Make sure that the parent object always has a field called '__systematics__'.
        """
        if "__systematics__" not in awkward.fields(self):
            self["__systematics__"] = {}

    @property
    def systematics(self):
        """
        Return the list of all systematics attached to this object.
        """
        regex = re.compile(r"\_{2}.*\_{2}")
        self._ensure_systematics()
        fields = [
            f for f in awkward.fields(self["__systematics__"]) if not regex.match(f)
        ]
        return self["__systematics__"][fields]

    @abstractmethod
    def _build_variations(
        self,
        name: str,
        what: Union[str, List[str], Tuple[str]],
        varying_function: Callable,
    ):
        """
        name: str, name of the systematic variation / uncertainty source
        what: Union[str, List[str], Tuple[str]], name what gets varied,
              this could be a list or tuple of column names
        varying_function: Union[function, bound method, partial], a function that describes how 'what' is varied
        define how to manipulate the output of varying_function to produce all systematic variations. Varying function
        must close over all non-event-data arguments.
        """
        pass

    @abstractmethod
    def explodes_how(self):
        """
        This describes how a systematic uncertainty needs to be evaluated in the context of other systematic uncertainties.
        i.e. Do you iterate over this keeping all others fixed or do you need to have correlations with other (subsets of) systematics.
        """
        # this function contains decades of thinking about iterate over systematics variations
        # your opinions about systematics go here. :D
        pass

    @abstractmethod
    def describe_variations(self):
        """returns a list of variation names"""
        pass

    def add_systematic(
        self,
        name: str,
        kind: str,
        what: Union[str, List[str], Tuple[str]],
        varying_function: Callable,
    ):
        """
        name: str, name of the systematic variation / uncertainty source
        kind: str, the name of the kind of systematic variation
        what: Union[str, List[str], Tuple[str]], name what gets varied,
               this could be a list or tuple of column names
        varying_function: Union[function, bound method], a function that describes how 'what' is varied, it must close over all non-event-data arguments.
        """
        self._ensure_systematics()

        if name in awkward.fields(self["__systematics__"]):
            raise ValueError(f"{name} already exists as a systematic for this object!")

        if kind not in self._systematic_kinds:
            raise ValueError(
                f"{kind} is not an available systematics type, please add it and try again!"
            )

        wrap = partial(
            awkward_rewrap, like_what=self["__systematics__"], gfunc=rewrap_recordarray
        )
        flat = (
            self
            if isinstance(self, coffea.nanoevents.methods.base.NanoEvents)
            else awkward.flatten(self)
        )

        if what == "weight" and "__ones__" not in awkward.fields(
            flat["__systematics__"]
        ):
            flat["__systematics__", "__ones__"] = numpy.ones(
                len(flat), dtype=numpy.float32
            )

        rendered_type = flat.layout.parameters["__record__"]
        as_syst_type = awkward.with_parameter(flat, "__record__", kind)
        as_syst_type._build_variations(name, what, varying_function)
        variations = as_syst_type.describe_variations()

        flat["__systematics__", name] = awkward.zip(
            {
                v: getattr(as_syst_type, v)(name, what, rendered_type)
                for v in variations
            },
            depth_limit=1,
            with_name=f"{name}Systematics",
        )

        self["__systematics__"] = wrap(flat["__systematics__"])
        self.behavior[("__typestr__", f"{name}Systematics")] = f"{kind}"


behavior[("__typestr__", "Systematic")] = "Systematic"

# initialize all systematic variation types
from coffea.nanoevents.methods import systematics

for kind in systematics.__all__:
    Systematic.add_kind(kind)


@awkward.mixin_class(behavior)
class NanoEvents(Systematic):
    """NanoEvents mixin class

    This mixin class is used as the top-level type for NanoEvents objects.
    """

    @property
    def metadata(self):
        """Arbitrary metadata"""
        return self.layout.purelist_parameter("metadata")


behavior[("__typestr__", "NanoEvents")] = "event"


@awkward.mixin_class(behavior)
class NanoCollection:
    """A NanoEvents collection

    This mixin provides some helper methods useful for creating cross-references
    and other advanced mixin types.
    """

    def _collection_name(self):
        """The name of the collection (i.e. the field under events where it is found)"""
        return self.layout.purelist_parameter("collection_name")

    def _getlistarray(self):
        """Do some digging to find the initial listarray"""

        def descend(layout, depth):
            islistarray = isinstance(
                layout,
                (awkward.layout.ListOffsetArray32, awkward.layout.ListOffsetArray64),
            )
            if islistarray and layout.content.parameter("collection_name") is not None:
                return lambda: layout

        return awkward._util.recursively_apply(self.layout, descend)

    def _content(self):
        """Internal method to get jagged collection content

        This should only be called on the original unsliced collection array.
        Used with global indexes to resolve cross-references"""
        return self._getlistarray().content

    def _apply_global_index(self, index):
        """Internal method to take from a collection using a flat index

        This is often necessary to be able to still resolve cross-references on
        reduced arrays or single records.
        """
        if isinstance(index, int):
            out = self._content()[index]
            return awkward.Record(out, behavior=self.behavior)

        def flat_take(layout):
            idx = awkward.Array(layout)
            return self._content()[idx.mask[idx >= 0]]

        def descend(layout, depth):
            if layout.purelist_depth == 1:
                return lambda: flat_take(layout)

        (index,) = awkward.broadcast_arrays(index)
        out = awkward._util.recursively_apply(index.layout, descend)
        return awkward.Array(out, behavior=self.behavior)

    def _events(self):
        """Internal method to get the originally-constructed NanoEvents

        This can be called at any time from any collection, as long as
        the NanoEventsFactory instance exists."""
        return self.behavior["__events_factory__"].events()


__all__ = ["NanoCollection", "NanoEvents", "Systematic"]



import awkward
import numpy
import os
import sys
import warnings

try:
    import cStringIO as io
except ImportError:
    import io

# for later
# func = numbaize(formula,['p%i'%i for i in range(nParms)]+[varnames[i] for i in range(nEvalVars)])


def is_gz_file(filename):
    with open(filename, "rb") as f:
        return f.read(2) == b"\x1f\x8b"


def _parse_jme_formatted_file(
    jmeFilePath, interpolatedFunc=False, parmsFromColumns=False, jme_f=None
):
    if jme_f is None:
        fopen = open
        fmode = "rt"
        if is_gz_file(jmeFilePath):
            import gzip

            fopen = gzip.open
            fmode = (
                "r"
                if sys.platform.startswith("win") and sys.version_info.major < 3
                else fmode
            )
        jme_f = fopen(jmeFilePath, fmode)
    layoutstr = jme_f.readline().strip().strip("{}")

    name = jmeFilePath.split("/")[-1].split(".")[0]

    layout = layoutstr.split()
    if not layout[0].isdigit():
        raise Exception("First column of JME File Header must be a digit!")

    # setup the file format
    nBinnedVars = int(layout[0])
    nBinColumns = 2 * nBinnedVars
    nEvalVars = int(layout[nBinnedVars + 1])
    formula = layout[nBinnedVars + nEvalVars + 2]
    nParms = 0
    while formula.count("[%i]" % nParms):
        formula = formula.replace("[%i]" % nParms, "p%i" % nParms)
        nParms += 1
    # get rid of TMath
    tmath = {
        "TMath::Max": "max",
        "TMath::Log": "log",
        "TMath::Power": "pow",
        "TMath::Erf": "erf",
    }
    for key, rpl in tmath.items():
        formula = formula.replace(key, rpl)
    # protect function names with vars in them
    funcs_to_cap = ["max", "exp", "pow"]

    # parse the columns
    minMax = ["Min", "Max"]
    columns = []
    dtypes = []
    offset = 1
    prefix = "" if parmsFromColumns else "binned_"
    for i in range(nBinnedVars):
        columns.extend(["%s%s%s" % (prefix, layout[i + offset], mm) for mm in minMax])
        dtypes.extend(["<f4", "<f4"])
    columns.append("NVars")
    dtypes.append("<i8")
    offset += nBinnedVars + 1
    prefix = "" if parmsFromColumns else "eval_"
    if not interpolatedFunc:
        for i in range(nEvalVars):
            columns.extend(
                ["%s%s%s" % (prefix, layout[i + offset], mm) for mm in minMax]
            )
            dtypes.extend(["<f4", "<f4"])
    for i in range(nParms):
        columns.append("p%i" % i)
        dtypes.append("<f4")

    for f in funcs_to_cap:
        formula = formula.replace(f, f.upper())

    templatevars = ["x", "y", "z", "t", "w", "s"]
    varnames = [layout[i + nBinnedVars + 2] for i in range(nEvalVars)]
    for find, replace in zip(templatevars, varnames):
        formula = formula.replace(find, replace.upper())
        funcs_to_cap.append(replace)
    # restore max
    for f in funcs_to_cap:
        formula = formula.replace(f.upper(), f)

    if parmsFromColumns:
        pars = numpy.genfromtxt(jme_f, encoding="ascii")
        if len(pars.shape) == 1:
            pars = pars[numpy.newaxis, :]
        nParms = pars.shape[1] - len(columns)
        for i in range(nParms):
            columns.append("p%i" % i)
            dtypes.append("<f4")
        pars = numpy.core.records.fromarrays(
            pars.transpose(), names=columns, formats=dtypes
        )
    else:
        pars = numpy.genfromtxt(
            jme_f,
            dtype=tuple(dtypes),
            names=tuple(columns),
            encoding="ascii",
        )
        if len(pars.shape) == 0:
            pars = pars[numpy.newaxis]

    outs = [
        name,
        layout,
        pars,
        nBinnedVars,
        nBinColumns,
        nEvalVars,
        formula,
        nParms,
        columns,
        dtypes,
    ]
    jme_f.close()
    return tuple(outs)


def _build_standard_jme_lookup(
    name,
    layout,
    pars,
    nBinnedVars,
    nBinColumns,
    nEvalVars,
    formula,
    nParms,
    columns,
    dtypes,
    interpolatedFunc=False,
):
    # the first bin is always usual for JECs
    # the next bins may vary in number, so they're jagged arrays... yay
    bins = {}
    offset_col = 0
    offset_name = 1
    bin_order = []
    for i in range(nBinnedVars):
        binMins = None
        binMaxs = None
        if i == 0:
            binMins = numpy.unique(pars[columns[0]])
            binMaxs = numpy.unique(pars[columns[1]])
            if numpy.all(binMins[1:] == binMaxs[:-1]):
                bins[layout[i + offset_name]] = numpy.union1d(binMins, binMaxs)
            else:
                warnings.warn(
                    "binning for file for %s is malformed in variable %s"
                    % (name, layout[i + offset_name])
                )
                bins[layout[i + offset_name]] = numpy.union1d(binMins, binMaxs[-1:])
        else:
            counts = numpy.zeros(0, dtype=numpy.int64)
            allBins = numpy.zeros(0, dtype=numpy.double)
            for binMin in bins[bin_order[0]][:-1]:
                binMins = numpy.unique(
                    pars[numpy.where(pars[columns[0]] == binMin)][
                        columns[i + offset_col]
                    ]
                )
                binMaxs = numpy.unique(
                    pars[numpy.where(pars[columns[0]] == binMin)][
                        columns[i + offset_col + 1]
                    ]
                )
                theBins = None
                if numpy.all(binMins[1:] == binMaxs[:-1]):
                    theBins = numpy.union1d(binMins, binMaxs)
                else:
                    warnings.warn(
                        "binning for file for %s is malformed in variable %s"
                        % (name, layout[i + offset_name])
                    )
                    theBins = numpy.union1d(binMins, binMaxs[-1:])
                allBins = numpy.append(allBins, theBins)
                counts = numpy.append(counts, theBins.size)
            bins[layout[i + offset_name]] = awkward.unflatten(allBins, counts)
        bin_order.append(layout[i + offset_name])
        offset_col += 1

    # skip nvars to the variable columns
    # the columns here define clamps for the variables defined in columns[]
    # ----> clamps can be different from bins
    # ----> if there is more than one binning variable this array is jagged
    # ----> just make it jagged all the time
    clamp_mins = {}
    clamp_maxs = {}
    var_order = []
    offset_col = 2 * nBinnedVars + 1
    offset_name = nBinnedVars + 2
    jagged_counts = numpy.ones(bins[bin_order[0]].size - 1, dtype=numpy.int64)
    if len(bin_order) > 1:
        jagged_counts = numpy.maximum(
            awkward.num(bins[bin_order[1]]) - 1, 0
        )  # need counts-1 since we only care about Nbins
    for i in range(nEvalVars):
        var_order.append(layout[i + offset_name])
        if not interpolatedFunc:
            clamp_mins[layout[i + offset_name]] = awkward.unflatten(
                numpy.atleast_1d(pars[columns[i + offset_col]]), jagged_counts
            )
            clamp_maxs[layout[i + offset_name]] = awkward.unflatten(
                numpy.atleast_1d(pars[columns[i + offset_col + 1]]), jagged_counts
            )
            assert awkward.is_valid(clamp_mins[layout[i + offset_name]])
            assert awkward.is_valid(clamp_maxs[layout[i + offset_name]])
            offset_col += 1

    # now get the parameters, which we will look up with the clamped values
    parms = []
    parm_order = []
    offset_col = 2 * nBinnedVars + 1 + int(not interpolatedFunc) * 2 * nEvalVars
    for i in range(nParms):
        jag = awkward.unflatten(pars[columns[i + offset_col]], jagged_counts)
        assert awkward.is_valid(jag)
        parms.append(jag)
        parm_order.append("p%i" % (i))

    wrapped_up = {}
    wrapped_up[(name, "jme_standard_function")] = (
        formula,
        (bins, bin_order),
        (clamp_mins, clamp_maxs, var_order),
        (parms, parm_order),
    )
    return wrapped_up


def _convert_standard_jme_txt_file(jmeFilePath):
    return _build_standard_jme_lookup(*_parse_jme_formatted_file(jmeFilePath))


convert_jec_txt_file = _convert_standard_jme_txt_file
convert_jr_txt_file = _convert_standard_jme_txt_file


def convert_jersf_txt_file(jersfFilePath):
    (
        name,
        layout,
        pars,
        nBinnedVars,
        nBinColumns,
        nEvalVars,
        formula,
        nParms,
        columns,
        dtypes,
    ) = _parse_jme_formatted_file(jersfFilePath, parmsFromColumns=True)

    temp = _build_standard_jme_lookup(
        name,
        layout,
        pars,
        nBinnedVars,
        nBinColumns,
        nEvalVars,
        formula,
        nParms,
        columns,
        dtypes,
    )
    wrapped_up = {}
    for key, val in temp.items():
        newkey = (key[0], "jersf_lookup")
        vallist = list(val)
        vals, names = vallist[-1]
        if len(vals) > 3:
            warnings.warn(
                "JERSF file is in the new format with split-out systematic, only parsing totals!!!"
            )
            vals = vals[:3]
        names = ["central-up-down"]
        central, down, up = vals
        vallist[-1] = ((central, up, down), names)
        wrapped_up[newkey] = tuple(vallist)

    return wrapped_up


def convert_junc_txt_file(juncFilePath):
    components = []
    basename = os.path.basename(juncFilePath).split(".")[0]
    fopen = open
    fmode = "rt"
    if is_gz_file(juncFilePath):
        import gzip

        fopen = gzip.open
        fmode = (
            "r"
            if sys.platform.startswith("win") and sys.version_info.major < 3
            else fmode
        )
    with fopen(juncFilePath, fmode) as uncfile:
        for line in uncfile:
            if line.startswith("#"):
                continue
            elif line.startswith("["):
                component_name = line.strip()[1:-1]  # remove leading and trailing []
                cname = "just/sum/dummy/dir/{0}_{1}.junc.txt".format(
                    basename, component_name
                )
                components.append((cname, []))
            elif components:
                components[-1][1].append(line)
            else:
                continue

    if not components:  # there are no components in the file
        components.append((juncFilePath, None))
    else:
        components = [(i, io.StringIO("".join(j))) for i, j in components]

    retval = {}
    for name, ifile in components:
        retval.update(convert_junc_txt_component(name, ifile))
    return retval


def convert_junc_txt_component(juncFilePath, uncFile):
    (
        name,
        layout,
        pars,
        nBinnedVars,
        nBinColumns,
        nEvalVars,
        formula,
        nParms,
        columns,
        dtypes,
    ) = _parse_jme_formatted_file(
        juncFilePath, interpolatedFunc=True, parmsFromColumns=True, jme_f=uncFile
    )

    temp = _build_standard_jme_lookup(
        name,
        layout,
        pars,
        nBinnedVars,
        nBinColumns,
        nEvalVars,
        formula,
        nParms,
        columns,
        dtypes,
        interpolatedFunc=True,
    )
    wrapped_up = {}
    for key, val in temp.items():
        newkey = (key[0], "jec_uncertainty_lookup")
        vallist = list(val)
        vals, names = vallist[-1]
        knots = vals[0 : len(vals) : 3]
        downs = vals[1 : len(vals) : 3]
        ups = vals[2 : len(vals) : 3]
        downs = numpy.array([numpy.array(awkward.flatten(down)) for down in downs])
        ups = numpy.array([numpy.array(awkward.flatten(up)) for up in ups])
        for knotv in knots:
            knot = numpy.unique(numpy.array(awkward.flatten(knotv)))
            if knot.size != 1:
                raise Exception("Multiple bin low edges found")
        knots = numpy.array(
            [numpy.unique(numpy.array(awkward.flatten(k)))[0] for k in knots]
        )
        vallist[2] = ({"knots": knots, "ups": ups.T, "downs": downs.T}, vallist[2][-1])
        vallist = vallist[:-1]
        wrapped_up[newkey] = tuple(vallist)
    return wrapped_up


def convert_effective_area_file(eaFilePath):
    fopen = open
    fmode = "rt"
    if is_gz_file(eaFilePath):
        import gzip

        fopen = gzip.open
        fmode = (
            "r"
            if sys.platform.startswith("win") and sys.version_info.major < 3
            else fmode
        )
    ea_f = fopen(eaFilePath, fmode)
    layoutstr = ea_f.readline().strip().strip("{}")
    ea_f.close()

    name = eaFilePath.split("/")[-1].split(".")[0]

    layout = layoutstr.split()
    if not layout[0].isdigit():
        raise Exception("First column of Effective Area File Header must be a digit!")

    # setup the file format
    nBinnedVars = int(layout[0])
    nEvalVars = int(layout[nBinnedVars + 1])

    minMax = ["Min", "Max"]
    columns = []
    dtypes = []
    offset = 1
    for i in range(nBinnedVars):
        columns.extend(["%s%s" % (layout[i + offset], mm) for mm in minMax])
        dtypes.extend(["<f4", "<f4"])
    offset += nBinnedVars + 1
    for i in range(nEvalVars):
        columns.append("%s" % (layout[i + offset]))
        dtypes.append("<f4")

    pars = numpy.genfromtxt(
        eaFilePath,
        dtype=tuple(dtypes),
        names=tuple(columns),
        skip_header=1,
        encoding="ascii",
    )

    bins = {}
    offset_col = 0
    offset_name = 1
    bin_order = []
    for i in range(nBinnedVars):
        binMins = None
        binMaxs = None
        if i == 0:
            binMins = numpy.unique(pars[columns[0]])
            binMaxs = numpy.unique(pars[columns[1]])
            bins[layout[i + offset_name]] = numpy.union1d(binMins, binMaxs)
        else:
            counts = numpy.zeros(0, dtype=numpy.int)
            allBins = numpy.zeros(0, dtype=numpy.double)
            for binMin in bins[bin_order[0]][:-1]:
                binMins = numpy.unique(
                    pars[numpy.where(pars[columns[0]] == binMin)][
                        columns[i + offset_col]
                    ]
                )
                binMaxs = numpy.unique(
                    pars[numpy.where(pars[columns[0]] == binMin)][
                        columns[i + offset_col + 1]
                    ]
                )
                theBins = numpy.union1d(binMins, binMaxs)
                allBins = numpy.append(allBins, theBins)
                counts = numpy.append(counts, theBins.size)
            bins[layout[i + offset_name]] = awkward.unflatten(allBins, counts)
        bin_order.append(layout[i + offset_name])
        offset_col += 1

    # again this is only for one dimension of binning, fight me
    # we can figure out a 2D EA when we get there
    offset_name += 1
    wrapped_up = {}
    lookup_type = "dense_lookup"
    dims = bins[layout[1]]
    for i in range(nEvalVars):
        ea_name = "_".join([name, columns[offset_name + i]])
        values = pars[columns[offset_name + i]]
        wrapped_up[(ea_name, lookup_type)] = (values, dims)

    return wrapped_up


def convert_rochester_file(path, loaduncs=True):
    initialized = False

    fopen = open
    fmode = "rt"
    if is_gz_file(path):
        import gzip

        fopen = gzip.open
        fmode = (
            "r"
            if sys.platform.startswith("win") and sys.version_info.major < 3
            else fmode
        )

    with fopen(path, fmode) as f:
        for line in f:
            line = line.strip("\n")
            # the number of sets available
            if line.startswith("NSET"):
                nsets = int(line.split()[1])
            # the number of members in a given set
            elif line.startswith("NMEM"):
                members = [int(x) for x in line.split()[1:]]
                assert len(members) == nsets
            # the type of the values provided: 0 is default, 1 is replicas (for stat unc), 2 is Symhes (for various systematics)
            elif line.startswith("TVAR"):
                tvars = [int(x) for x in line.split()[1:]]
                assert len(tvars) == nsets
            # number of phi bins
            elif line.startswith("CPHI"):
                nphi = int(line.split()[1])
                phiedges = [
                    float(x) * 2 * numpy.pi / nphi - numpy.pi for x in range(nphi + 1)
                ]
            # number of eta bins and edges
            elif line.startswith("CETA"):
                neta = int(line.split()[1])
                etaedges = [float(x) for x in line.split()[2:]]
                assert len(etaedges) == neta + 1
            # minimum number of tracker layers with measurement
            elif line.startswith("RMIN"):
                nmin = int(line.split()[1])
            # number of bins in the number of tracker layers measurements
            elif line.startswith("RTRK"):
                ntrk = int(line.split()[1])
            # number of abseta bins and edges
            elif line.startswith("RETA"):
                nabseta = int(line.split()[1])
                absetaedges = [float(x) for x in line.split()[2:]]
                assert len(absetaedges) == nabseta + 1
            # load the parameters
            # the structure will be
            # SETNUMBER MEMBERNUMBER TAG[T,R,F,C] [TAG specific indices] [values]
            else:
                if not initialized:
                    # don't want to necessarily load uncertainties
                    toload = nsets if loaduncs else 1
                    M = {
                        s: {m: {t: {} for t in range(2)} for m in range(members[s])}
                        for s in range(toload)
                    }
                    A = {
                        s: {m: {t: {} for t in range(2)} for m in range(members[s])}
                        for s in range(toload)
                    }
                    kRes = {
                        s: {m: {t: [] for t in range(2)} for m in range(members[s])}
                        for s in range(toload)
                    }
                    rsPars = {
                        s: {m: {t: {} for t in range(3)} for m in range(members[s])}
                        for s in range(toload)
                    }
                    cbS = {s: {m: {} for m in range(members[s])} for s in range(toload)}
                    cbA = {s: {m: {} for m in range(members[s])} for s in range(toload)}
                    cbN = {s: {m: {} for m in range(members[s])} for s in range(toload)}
                    initialized = True
                remainder = line.split()
                setn, membern, tag, *remainder = remainder
                setn = int(setn)
                membern = int(membern)
                tag = str(tag)
                # tag T has 2 indices corresponding to TYPE BINNUMBER and has RTRK+1 values each
                # these correspond to the nTrk[2] parameters of RocRes (and BINNUMBER is the abseta bin)
                if tag == "T":
                    t, b, *remainder = remainder
                    t = int(t)
                    b = int(b)
                    values = [float(x) for x in remainder]
                    assert len(values) == ntrk + 1

                # tag R has 2 indices corresponding to VARIABLE BINNUMBER and has RTRK values each
                # these variables correspond to the rsPar[3] and crystal ball (std::vector<CrystalBall> cb) of RocRes where CrystalBall has valus s, a, n
                # (and BINNUMBER is the abseta bin)
                # Note: crystal ball here is a symmetric double-sided crystal ball
                elif tag == "R":
                    v, b, *remainder = remainder
                    v = int(v)
                    b = int(b)
                    values = [float(x) for x in remainder]
                    assert len(values) == ntrk
                    if v in range(3):
                        if setn in rsPars:
                            rsPars[setn][membern][v][b] = values
                            if v == 2:
                                rsPars[setn][membern][v][b] = [x * 0.01 for x in values]
                    elif v == 3:
                        if setn in cbS:
                            cbS[setn][membern][b] = values
                    elif v == 4:
                        if setn in cbA:
                            cbA[setn][membern][b] = values
                    elif v == 5:
                        if setn in cbN:
                            cbN[setn][membern][b] = values

                # tag F has 1 index corresponding to TYPE and has RETA values each
                # these correspond to the kRes[2] of RocRes
                elif tag == "F":
                    t, *remainder = remainder
                    t = int(t)
                    values = [float(x) for x in remainder]
                    assert len(values) == nabseta
                    if setn in kRes:
                        kRes[setn][membern][t] = values

                # tag C has 3 indices corresponding to TYPE VARIABLE BINNUMBER and has NPHI values each
                # these correspond to M and A values of CorParams (and BINNUMBER is the eta bin)
                # These are what are used to get the scale factor for kScaleDT (and kScaleMC)
                #       scale = 1.0 / (M+Q*A*pT)
                # For the kSpreadMC (gen matched, recommended) and kSmearMC (not gen matched), we need all of the above parameters
                elif tag == "C":
                    t, v, b, *remainder = remainder
                    t = int(t)
                    v = int(v)
                    b = int(b)
                    values = [float(x) for x in remainder]
                    assert len(values) == nphi
                    if v == 0:
                        if setn in M:
                            M[setn][membern][t][b] = [1.0 + x * 0.01 for x in values]
                    elif v == 1:
                        if setn in A:
                            A[setn][membern][t][b] = [x * 0.01 for x in values]

                else:
                    raise ValueError(line)

    # now build the lookup tables
    # for data scale, simple, just M A in bins of eta,phi
    _scaleedges = (numpy.array(etaedges), numpy.array(phiedges))
    _Mvalues = {
        s: {
            m: {t: numpy.array([M[s][m][t][b] for b in range(neta)]) for t in M[s][m]}
            for m in M[s]
        }
        for s in M
    }
    _Avalues = {
        s: {
            m: {t: numpy.array([A[s][m][t][b] for b in range(neta)]) for t in A[s][m]}
            for m in A[s]
        }
        for s in A
    }

    # for mc scale, more complicated
    # version 1 if gen pt available
    # only requires the kRes lookup
    _resedges = numpy.array(absetaedges)
    _kResvalues = {
        s: {m: {t: numpy.array(kRes[s][m][t]) for t in kRes[s][m]} for m in kRes[s]}
        for s in kRes
    }

    # version 2 if gen pt not available
    trkedges = [0] + [nmin + x + 0.5 for x in range(ntrk)]
    _cbedges = (numpy.array(absetaedges), numpy.array(trkedges))
    _rsParsvalues = {
        s: {
            m: {
                t: numpy.array([rsPars[s][m][t][b] for b in range(nabseta)])
                for t in rsPars[s][m]
            }
            for m in rsPars[s]
        }
        for s in rsPars
    }
    _cbSvalues = {
        s: {m: numpy.array([cbS[s][m][b] for b in range(nabseta)]) for m in cbS[s]}
        for s in cbS
    }
    _cbAvalues = {
        s: {m: numpy.array([cbA[s][m][b] for b in range(nabseta)]) for m in cbA[s]}
        for s in cbA
    }
    _cbNvalues = {
        s: {m: numpy.array([cbN[s][m][b] for b in range(nabseta)]) for m in cbN[s]}
        for s in cbN
    }

    wrapped_up = {
        "nsets": nsets,
        "members": members,
        "edges": {
            "scales": _scaleedges,
            "res": _resedges,
            "cb": _cbedges,
        },
        "values": {
            "M": _Mvalues,
            "A": _Avalues,
            "kRes": _kResvalues,
            "rsPars": _rsParsvalues,
            "cbS": _cbSvalues,
            "cbA": _cbAvalues,
            "cbN": _cbNvalues,
        },
    }
    return wrapped_up



"""NanoEvents and helpers

"""
from coffea.nanoevents.factory import NanoEventsFactory
from coffea.nanoevents.schemas import (
    BaseSchema,
    NanoAODSchema,
    PFNanoAODSchema,
    TreeMakerSchema,
    PHYSLITESchema,
    DelphesSchema,
    PDUNESchema,
)

__all__ = [
    "NanoEventsFactory",
    "BaseSchema",
    "NanoAODSchema",
    "PFNanoAODSchema",
    "TreeMakerSchema",
    "PHYSLITESchema",
    "DelphesSchema",
    "PDUNESchema",
]



"""Utility functions

"""
from typing import List, Optional, Any
import awkward
import hashlib
import numpy
import numba
import coffea
from rich.progress import (
    Progress,
    BarColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
    Column,
    ProgressColumn,
    Text,
)

ak = awkward
np = numpy
nb = numba

import lz4.frame
import cloudpickle
import warnings
from functools import partial


def load(filename):
    """Load a coffea file from disk"""
    with lz4.frame.open(filename) as fin:
        output = cloudpickle.load(fin)
    return output


def save(output, filename):
    """Save a coffea object or collection thereof to disk

    This function can accept any picklable object.  Suggested suffix: ``.coffea``
    """
    with lz4.frame.open(filename, "wb") as fout:
        thepickle = cloudpickle.dumps(output)
        fout.write(thepickle)


def _hex(string):
    try:
        return string.hex()
    except AttributeError:
        return "".join("{:02x}".format(ord(c)) for c in string)


def _ascii(maybebytes):
    try:
        return maybebytes.decode("ascii")
    except AttributeError:
        return maybebytes


def _hash(items):
    # python 3.3 salts hash(), we want it to persist across processes
    x = hashlib.md5(bytes(";".join(str(x) for x in items), "ascii"))
    return int(x.hexdigest()[:16], base=16)


def _ensure_flat(array, allow_missing=False):
    """Normalize an array to a flat numpy array or raise ValueError"""
    if not isinstance(array, (ak.Array, numpy.ndarray)):
        raise ValueError("Expected a numpy or awkward array, received: %r" % array)

    aktype = ak.type(array)
    if not isinstance(aktype, ak.types.ArrayType):
        raise ValueError("Expected an array type, received: %r" % aktype)
    isprimitive = isinstance(aktype.type, ak.types.PrimitiveType)
    isoptionprimitive = isinstance(aktype.type, ak.types.OptionType) and isinstance(
        aktype.type.type, ak.types.PrimitiveType
    )
    if allow_missing and not (isprimitive or isoptionprimitive):
        raise ValueError(
            "Expected an array of type N * primitive or N * ?primitive, received: %r"
            % aktype
        )
    if not (allow_missing or isprimitive):
        raise ValueError(
            "Expected an array of type N * primitive, received: %r" % aktype
        )
    if isinstance(array, ak.Array):
        array = ak.to_numpy(array, allow_missing=allow_missing)
    return array


def _exception_chain(exc: BaseException) -> List[BaseException]:
    """Retrieves the entire exception chain as a list."""
    ret = []
    while isinstance(exc, BaseException):
        ret.append(exc)
        exc = exc.__cause__
    return ret


class SpeedColumn(ProgressColumn):
    """Renders human readable transfer speed."""

    def __init__(self, fmt: str = ".1f", table_column: Optional[Column] = None):
        self.fmt = fmt
        super().__init__(table_column=table_column)

    def render(self, task: Any) -> Text:
        """Show data transfer speed."""
        speed = task.finished_speed or task.speed
        if speed is None:
            return Text("?", style="progress.data.speed")
        return Text(f"{speed:{self.fmt}}", style="progress.data.speed")


def rich_bar():
    return Progress(
        TextColumn("[bold blue]{task.description}", justify="right"),
        "[progress.percentage]{task.percentage:>3.0f}%",
        BarColumn(bar_width=None),
        TextColumn(
            "[bold blue][progress.completed]{task.completed}/{task.total}",
            justify="right",
        ),
        "[",
        TimeElapsedColumn(),
        "<",
        TimeRemainingColumn(),
        "|",
        SpeedColumn(".1f"),
        TextColumn("[progress.data.speed]{task.fields[unit]}/s", justify="right"),
        "]",
        auto_refresh=False,
    )


# lifted from awkward - https://github.com/scikit-hep/awkward-1.0/blob/5fe31a916bf30df6c2ea10d4094f6f1aefcf3d0c/src/awkward/_util.py#L47-L61 # noqa
# drive our deprecations-as-errors as with awkward
def deprecate(exception, version, date=None):
    if coffea.deprecations_as_errors:
        raise exception
    else:
        if date is None:
            date = ""
        else:
            date = " (target date: " + date + ")"
        message = """In coffea version {0}{1}, this will be an error.
(Set coffea.deprecations_as_errors = True to get a stack trace now.)
{2}: {3}""".format(
            version, date, type(exception).__name__, str(exception)
        )
        warnings.warn(message, FutureWarning)


# re-nest a record array into a ListArray
def awkward_rewrap(arr, like_what, gfunc):
    behavior = awkward._util.behaviorof(like_what)
    func = partial(gfunc, data=arr.layout)
    layout = awkward.operations.convert.to_layout(like_what)
    newlayout = awkward._util.recursively_apply(layout, func)
    return awkward._util.wrap(newlayout, behavior=behavior)


# we're gonna assume that the first record array we encounter is the flattened data
def rewrap_recordarray(layout, depth, data):
    if isinstance(layout, awkward.layout.RecordArray):
        return lambda: data
    return None



import warnings
import json
from coffea.nanoevents.mapping.base import UUIDOpener, BaseSourceMapping
from coffea.nanoevents.util import quote, tuple_to_key


class SimplePreloadedColumnSource(dict):
    def __init__(self, columns, uuid, num_rows, object_path, **kwargs):
        self.update(columns)
        self.metadata = {"uuid": uuid, "num_rows": num_rows, "object_path": object_path}
        self.metadata.update(kwargs)


class PreloadedOpener(UUIDOpener):
    def __init__(self, uuid_pfnmap):
        super(PreloadedOpener, self).__init__(uuid_pfnmap)

    def open_uuid(self, uuid):
        pcs = self._uuid_pfnmap[uuid]
        if str(pcs.uuid) != uuid:
            raise RuntimeError(
                f"UUID of array source {pcs} does not match expected value ({uuid})"
            )
        return pcs


class PreloadedSourceMapping(BaseSourceMapping):
    _debug = False

    def __init__(self, array_source, cache=None, access_log=None):
        super(PreloadedSourceMapping, self).__init__(array_source, cache, access_log)

    @classmethod
    def _extract_base_form(cls, column_source):
        branch_forms = {}
        for key, branch in column_source.items():
            if "," in key or "!" in key:
                warnings.warn(
                    f"Skipping {key} because it contains characters that NanoEvents cannot accept [,!]"
                )
                continue
            form = json.loads(branch.layout.form.tojson())
            if (
                form["class"].startswith("ListOffset")
                and form["content"]["class"] == "NumpyArray"  # noqa
            ):
                form["form_key"] = quote(f"{key},!load")
                form["content"]["form_key"] = quote(f"{key},!load,!content")
                form["content"]["parameters"] = {"__doc__": key}
            elif form["class"] == "NumpyArray":
                form["form_key"] = quote(f"{key},!load")
                form["parameters"] = {"__doc__": key}
            else:
                warnings.warn(
                    f"Skipping {key} as it is not interpretable by NanoEvents"
                )
                continue
            branch_forms[key] = form

        return {
            "class": "RecordArray",
            "contents": branch_forms,
            "parameters": {"__doc__": "preloaded column source"},
            "form_key": "",
        }

    def key_root(self):
        return "PreloadedSourceMapping:"

    def preload_column_source(self, uuid, path_in_source, source):
        """To save a double-open when using NanoEventsFactory.from_file"""
        key = self.key_root() + tuple_to_key((uuid, path_in_source))
        self._cache[key] = source

    def get_column_handle(self, columnsource, name):
        return columnsource[name]

    def extract_column(self, columnhandle, start, stop):
        # make sure uproot is single-core since our calling context might not be
        return columnhandle[start:stop]

    def __len__(self):
        raise NotImplementedError

    def __iter__(self):
        raise NotImplementedError



from functools import wraps


def timeout(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        import signal

        def _timeout_handler(signum, frame):
            raise Exception("Timeout hit")

        signal.signal(signal.SIGALRM, _timeout_handler)
        if kwargs.get("timeout"):
            signal.alarm(max(1, int(kwargs["timeout"])))
        try:
            result = func(*args, **kwargs)
        finally:
            signal.alarm(0)
        return result

    return wrapper



import warnings
import uproot
import awkward
import json
from coffea.nanoevents.mapping.base import UUIDOpener, BaseSourceMapping
from coffea.nanoevents.util import quote, tuple_to_key


class TrivialUprootOpener(UUIDOpener):
    def __init__(self, uuid_pfnmap, uproot_options={}):
        super(TrivialUprootOpener, self).__init__(uuid_pfnmap)
        self._uproot_options = uproot_options

    def open_uuid(self, uuid):
        pfn = self._uuid_pfnmap[uuid]
        rootdir = uproot.open(pfn, **self._uproot_options)
        if str(rootdir.file.uuid) != uuid:
            raise RuntimeError(
                f"UUID of file {pfn} does not match expected value ({uuid})"
            )
        return rootdir


class CannotBeNanoEvents(Exception):
    pass


def _lazify_form(form, prefix, docstr=None):
    if not isinstance(form, dict) or "class" not in form:
        raise RuntimeError("form should have been normalized by now")
    elif form["class"].startswith("ListOffset"):
        # awkward will add !offsets
        form["form_key"] = quote(prefix)
        form["content"] = _lazify_form(
            form["content"], prefix + ",!content", docstr=docstr
        )
    elif form["class"] == "NumpyArray":
        form["form_key"] = quote(prefix)
        if docstr is not None:
            form["parameters"] = {"__doc__": docstr}
    elif form["class"] == "RegularArray":
        form["content"] = _lazify_form(
            form["content"], prefix + ",!content", docstr=docstr
        )
        if docstr is not None:
            form["parameters"] = {"__doc__": docstr}
    elif form["class"] == "RecordArray":
        for field in list(form["contents"]):
            if "," in field or "!" in field:
                raise CannotBeNanoEvents(
                    f"A subform contains a field with invalid characters: {field}"
                )
            form["contents"][field] = _lazify_form(
                form["contents"][field], prefix + f",{field},!item"
            )
        if docstr is not None:
            form["parameters"] = {"__doc__": docstr}
    else:
        raise CannotBeNanoEvents("Unknown form")
    return form


class UprootSourceMapping(BaseSourceMapping):
    _debug = False
    _fix_awkward_form_of_iter = True

    def __init__(self, fileopener, cache=None, access_log=None):
        super(UprootSourceMapping, self).__init__(fileopener, cache, access_log)

    @classmethod
    def _extract_base_form(cls, tree, iteritems_options={}):
        branch_forms = {}
        for key, branch in tree.iteritems(**iteritems_options):
            if key in branch_forms:
                warnings.warn(
                    f"Found duplicate branch {key} in {tree}, taking first instance"
                )
                continue
            if "," in key or "!" in key:
                warnings.warn(
                    f"Skipping {key} because it contains characters that NanoEvents cannot accept [,!]"
                )
                continue
            if len(branch):
                # The branch is split and its sub-branches will be enumerated by tree.iteritems
                continue
            if isinstance(
                branch.interpretation,
                uproot.interpretation.identify.UnknownInterpretation,
            ):
                warnings.warn(f"Skipping {key} as it is not interpretable by Uproot")
                continue
            try:
                form = branch.interpretation.awkward_form(None)
            except uproot.interpretation.objects.CannotBeAwkward:
                warnings.warn(
                    f"Skipping {key} as it is it cannot be represented as an Awkward array"
                )
                continue
            # until awkward-forth is available, this fixer is necessary
            if cls._fix_awkward_form_of_iter:
                form = uproot._util.recursively_fix_awkward_form_of_iter(
                    awkward, branch.interpretation, form
                )
            form = uproot._util.awkward_form_remove_uproot(awkward, form)
            form = json.loads(
                form.tojson()
            )  # normalizes form (expand NumpyArray classes)
            try:
                form = _lazify_form(form, f"{key},!load", docstr=branch.title)
            except CannotBeNanoEvents as ex:
                warnings.warn(
                    f"Skipping {key} as it is not interpretable by NanoEvents\nDetails: {ex}"
                )
                continue
            branch_forms[key] = form

        return {
            "class": "RecordArray",
            "contents": branch_forms,
            "parameters": {"__doc__": tree.title},
            "form_key": "",
        }

    def key_root(self):
        return "UprootSourceMapping:"

    def preload_column_source(self, uuid, path_in_source, source):
        """To save a double-open when using NanoEventsFactory.from_file"""
        key = self.key_root() + tuple_to_key((uuid, path_in_source))
        self._cache[key] = source

    def get_column_handle(self, columnsource, name):
        return columnsource[name]

    def extract_column(self, columnhandle, start, stop):
        # make sure uproot is single-core since our calling context might not be
        return columnhandle.array(
            entry_start=start,
            entry_stop=stop,
            decompression_executor=uproot.source.futures.TrivialExecutor(),
            interpretation_executor=uproot.source.futures.TrivialExecutor(),
        )

    def __len__(self):
        raise NotImplementedError

    def __iter__(self):
        raise NotImplementedError



"""Tools to parse CMS luminosity non-event data

These tools are currently tailored to the CMS experiment
data formats, however they could be generalized and/or compartmentalized
into a standalone package.
"""
from .lumi_tools import (
    LumiData,
    LumiList,
    LumiMask,
)

__all__ = [
    "LumiData",
    "LumiList",
    "LumiMask",
]



from coffea.lookup_tools.lookup_base import lookup_base

import numpy
import numba

from copy import deepcopy


# methods for dealing with b-tag SFs
@numba.jit(forceobj=True)
def numba_apply_1d(functions, variables):
    out = numpy.empty(variables.shape)
    for i in range(functions.size):
        out[i] = functions[i](variables[i])
    return out


def numbaize(fstr, varlist):
    """
    Convert function string to numba function
    Supports only simple math for now
    """

    lstr = "lambda %s: %s" % (",".join(varlist), fstr)
    func = eval(lstr, {"log": numpy.log, "sqrt": numpy.sqrt})
    nfunc = numba.njit(func)
    return nfunc


# methods for dealing with b-tag SFs
class dense_evaluated_lookup(lookup_base):
    def __init__(self, values, dims, feval_dim=None):
        super(dense_evaluated_lookup, self).__init__()
        self._dimension = 0
        whattype = type(dims)
        if whattype == numpy.ndarray:
            self._dimension = 1
        else:
            self._dimension = len(dims)
        if self._dimension == 0:
            raise Exception("Could not define dimension for {}".format(whattype))
        self._axes = deepcopy(dims)
        self._feval_dim = None
        vals_are_strings = (
            "string" in values.dtype.name
            or "str" in values.dtype.name
            or "unicode" in values.dtype.name
            or "bytes" in values.dtype.name
        )  # ....
        if not isinstance(values, numpy.ndarray):
            raise TypeError("values is not a numpy array, but %r" % type(values))
        if not vals_are_strings:
            raise Exception("Non-string values passed to dense_evaluated_lookup!")
        if feval_dim is None:
            raise Exception(
                "Evaluation dimensions not specified in dense_evaluated_lookup"
            )
        funcs = numpy.zeros(shape=values.shape, dtype="O")
        for i in range(values.size):
            idx = numpy.unravel_index(i, shape=values.shape)
            funcs[idx] = numbaize(values[idx], ["x"])
        self._values = deepcopy(funcs)
        # TODO: support for multidimensional functions and functions with variables other than 'x'
        if len(feval_dim) > 1:
            raise Exception(
                "lookup_tools.evaluator only accepts 1D functions right now!"
            )
        self._feval_dim = feval_dim[0]

    def _evaluate(self, *args):
        indices = []
        if self._dimension == 1:
            indices.append(
                numpy.clip(
                    numpy.searchsorted(self._axes, args[0], side="right") - 1,
                    0,
                    self._values.shape[0] - 1,
                )
            )
        else:
            for dim in range(self._dimension):
                indices.append(
                    numpy.clip(
                        numpy.searchsorted(self._axes[dim], args[dim], side="right")
                        - 1,
                        0,
                        self._values.shape[len(self._axes) - dim - 1] - 1,
                    )
                )
        indices.reverse()
        return numba_apply_1d(self._values[tuple(indices)], args[self._feval_dim])

    def __repr__(self):
        myrepr = "{} dimensional histogram with axes:\n".format(self._dimension)
        temp = ""
        if self._dimension == 1:
            temp = "\t1: {}\n".format(self._axes)
        else:
            temp = "\t1: {}\n".format(self._axes[0])
        for idim in range(1, self._dimension):
            temp += "\t{}: {}\n".format(idim + 1, self._axes[idim])
        myrepr += temp
        return myrepr



# import warnings
# from collections import defaultdict
# import copy
from coffea.nanoevents.schemas.base import BaseSchema, zip_forms

# from coffea.nanoevents.util import quote
import collections.abc
import fnmatch


class PDUNESchema(BaseSchema):
    mixins = {
        # branch tag : obj. class
        "RecoBeam": "Beam",
        "Tracks": "Tracks",
        "Showers": "Showers",
        "reco_beam": "RecoBeam",
        "reco_daughter_allTrack": "Tracks",
        "reco_daughter_allShower": "Showers",
        "start3D": "ThreeVector",
        "end3D": "ThreeVector",
        "start4D": "LorentzVector",
        "end4D": "LorentzVector",
        "vtx3D": "ThreeVector",
    }

    top_objects = {
        "reco_beam": "RecoBeam",
        "reco_daughter_allTrack": "Tracks",
        "reco_daughter_allShower": "Showers",
        "true_beam": "TrueBeam",
    }

    def __init__(self, base_form):
        super().__init__(base_form)
        # print(self.branch_behavior_dict)
        self._form["contents"] = self._build_collections(self._form["contents"])

    # build a dictionary of hierarchy i.e.
    # dict["RecoBeam"]["start"]["x/y/z"]=form
    def _recursion(self, key_list, obj, key_dict, i):
        if i < len(key_list) - 1:  # not the last element
            curr_key = key_list[i]
            if curr_key not in key_dict.keys():
                key_dict[curr_key] = {}
            self._recursion(key_list, obj, key_dict[curr_key], i + 1)
        else:
            if i > 0:
                if key_dict.get(key_list[i]) is None:
                    key_dict[key_list[i]] = obj

    # RecoBeam--> start/end/len --> x,y,z
    def _recursive_zip(self, forms, hierarchy, key, final_zip=False):
        # print(key,'\n')
        for k, v in hierarchy.items():
            if isinstance(v, collections.abc.Mapping):
                # print('if:', k,v)
                forms[k] = self._recursive_zip(
                    forms.get(k, {}), hierarchy.get(k, {}), k, True
                )
                # form = zip_forms(form,"events",record_name=None)
            else:
                name = self.mixins[k] if k in self.mixins.keys() else None
                # print('else:', k,v,name)
                forms[k] = zip_forms(forms[k], k, record_name=name)
        if final_zip:
            forms = zip_forms(forms, key, record_name=None)
        return forms

    # forms --> forms
    # first level
    # k,v -> RecoBeam, map
    # second level
    #   forms --> forms[RecoBeam]
    #   k, v -> start:obj
    #   goto else
    #   forms[RecoBeam][start]=zip_forms
    #   forms[RecoBeam][end]=zip_forms
    #   then needs to zip forms[RecoBeam]

    def _filter_branches(self, branches, wildcard):
        return [b for b in branches if fnmatch.fnmatch(b, wildcard)]

    def _type_dictionary_builder(self, branch_forms):
        all_branches = branch_forms.keys()

        # 1. Build ThreeVector Suffix
        V3Var = ["X", "Y", "Z"]
        V4Var = ["Px", "Py", "Pz", "E"]
        # get suffix, check if it ends with X,Y, or Z
        V3Sets = [
            set(
                [
                    b.split("_")[-1][:-1]
                    for b in self._filter_branches(all_branches, "*%s" % V)
                ]
            )
            for V in V3Var
        ]
        V3Set = V3Sets[0].intersection(V3Sets[1], V3Sets[2])

        V4Sets = [
            set(
                [
                    b.split("_")[-1][: -len(V)]
                    for b in self._filter_branches(all_branches, "*%s" % V)
                ]
            )
            for V in V4Var
        ]
        V4Set = V4Sets[0].intersection(V4Sets[1], V4Sets[2])

        V3Names = [s + v for s in V3Set for v in V3Var]
        V4Names = [s + v for s in V4Set for v in V4Var]
        # note v4 is a subset of v3

        branch_behavior_dict = {}
        # print(V3Set, V4Set)
        for b in all_branches:
            b_fields = b.split("_")
            b_end = b_fields[-1]

            behavior = ""
            if b_end in V3Names:
                behavior = "ThreeVector"
            elif b_end in V4Names:
                behavior = "FourVector"
            branch_behavior_dict[b] = behavior

        return branch_behavior_dict

    def _sort_branches(self, branches):
        sorted_branches = sorted(
            branches,
            key=lambda x: len(x.split("_")) + (self.branch_behavior_dict[x] != ""),
            reverse=True,
        )
        return sorted_branches

    def _build_collections(self, branch_forms):
        self.branch_behavior_dict = self._type_dictionary_builder(branch_forms)

        key_form_dict = (
            {}
        )  # finally formed nested dictionary, the input to build branch form structure
        key_dict = (
            {}
        )  # instead of having `form` in the leaf, using the string 'ak_form'

        # i.e. reco_beam.startPoint --> reco_beam, startPoint
        # branch_keys = {}  # branch_name:[keys, pos]--> a data struct to store

        obj_lists = list(self.top_objects.keys())

        branches = self._sort_branches(
            branch_forms.keys()
        )  # sorted branches used to create nanoevent

        for key in branches:
            ak_form = branch_forms[key]

            behavior = self.branch_behavior_dict[key]

            which_top_key = list(map(lambda v: v in key, obj_lists))
            if sum(which_top_key) == 0:
                continue

            top_key = "".join([t * w for t, w in zip(which_top_key, obj_lists)])
            objname = self.top_objects[top_key]
            sub_keys = key.replace(top_key, objname).split("_")[1:]
            last_key = sub_keys[-1]
            # modify last key to X/Y/Z/Px/Py/Pz/E if this is 3D or 4D object
            if behavior != "":
                v = None
                if any(last_key.endswith(v) for v in ["X", "Y", "Z", "E"]):
                    v = last_key[-1].lower()
                    v = "energy" if v == "e" else v
                    last_key = last_key[:-1] + ("3D" if v != "energy" else "4D")
                if any(last_key.endswith(v) for v in ["Px", "Py", "Pz"]):
                    v = last_key[-2].lower()
                    last_key = last_key[:-1] + "4D"
                sub_keys[-1] = last_key
                self.mixins[last_key] = (
                    "ThreeVector" if last_key.endswith("3D") else "LorentzVector"
                )
                sub_keys.append(v)

            # sub_key = "_".join(sub_keys)

            keys = [objname] + sub_keys

            self._recursion(
                keys, ak_form, key_form_dict, 0
            )  # add form to end of dictionary

            self._recursion(
                keys[:-1], "obj", key_dict, 0
            )  # add string "obj" to end of dictionary

            # print(ak_form)

        # print(key_form_dict)
        # print(key_form_dict)
        # print(key_dict)
        # print(key_form_dict)

        # print(key_form_dict,'\n')
        # print(key_dict,'\n')

        output = self._recursive_zip(key_form_dict, key_dict, "Events")

        return output

    # @staticmethod
    # def _create_eventindex_form(base_form, key):
    #    form = copy.deepcopy(base_form)
    #    form["content"] = {
    #        "class": "NumpyArray",
    #        "parameters": {},
    #        "form_key": quote(f"{key},!load,!eventindex,!content"),
    #        "itemsize": 8,
    #        "primitive": "int64",
    #    }
    #    return form

    @property
    def behavior(self):
        """Behaviors necessary to implement this schema"""
        from coffea.nanoevents.methods import pdune

        return pdune.behavior



from ..util import numpy
import correctionlib
import json


def is_gz_file(filename):
    with open(filename, "rb") as f:
        return f.read(2) == b"\x1f\x8b"


def extract_json_histo_structure(parselevel, axis_names, axes):
    if "value" in parselevel.keys():
        return
    name = list(parselevel)[0].split(":")[0]
    bins_pairs = [
        key.split(":")[-1].strip("[]").split(",") for key in parselevel.keys()
    ]
    bins = []
    for pair in bins_pairs:
        bins.extend([float(val) for val in pair])
    bins.sort()
    bins = numpy.unique(numpy.array(bins))
    axis_names.append(name.encode())
    axes[axis_names[-1]] = bins
    extract_json_histo_structure(parselevel[list(parselevel)[0]], axis_names, axes)


def extract_json_histo_values(parselevel, binlows, values, val_names):
    if "value" in parselevel.keys():
        binvals = {}
        binvals.update(parselevel)
        keylows = tuple(binlows)
        values[keylows] = binvals
        for key in parselevel.keys():
            val_names.add(key)
        return
    for key in parselevel.keys():
        lowside = float(key.split(":")[-1].strip("[]").split(",")[0])
        thelows = [lowside]
        if len(binlows) != 0:
            thelows = binlows + thelows
        extract_json_histo_values(parselevel[key], thelows, values, val_names)


def convert_histo_json_file(filename):
    file = open(filename)
    info = json.load(file)
    file.close()
    names_and_orders = {}
    names_and_axes = {}
    names_and_binvalues = {}
    names_and_valnames = {}

    # first pass, convert info['dir']['hist_title'] to dir/hist_title
    # and un-nest everything from the json structure, make binnings, etc.
    for dir in info.keys():
        for htitle in info[dir].keys():
            axis_order = []  # keep the axis order
            axes = {}
            bins_and_values = {}
            val_names = set()
            extract_json_histo_structure(info[dir][htitle], axis_order, axes)
            extract_json_histo_values(info[dir][htitle], [], bins_and_values, val_names)
            histname = "%s/%s" % (dir, htitle)
            names_and_axes[histname] = axes
            names_and_orders[histname] = axis_order
            names_and_binvalues[histname] = bins_and_values
            names_and_valnames[histname] = val_names

    wrapped_up = {}
    for name, axes in names_and_axes.items():
        theshape = tuple([axes[axis].size - 1 for axis in names_and_orders[name]])
        valsdict = {}
        for vname in names_and_valnames[histname]:
            valsdict[vname] = numpy.zeros(shape=theshape).flatten()
        flatidx = numpy.arange(numpy.zeros(shape=theshape).size)
        binidx = numpy.unravel_index(flatidx, shape=theshape)
        for vname in valsdict:
            for iflat in flatidx:
                binlows = []
                for idim, axis in enumerate(names_and_orders[name]):
                    binlows.append(axes[axis][binidx[idim][iflat]])
                thevals = names_and_binvalues[name][tuple(binlows)]
                valsdict[vname][iflat] = thevals[vname]
            valsdict[vname] = valsdict[vname].reshape(theshape)
        bins_in_order = []
        for axis in names_and_orders[name]:
            bins_in_order.append(axes[axis])
        for vname in valsdict:
            wrapped_up[(name + "_" + vname, "dense_lookup")] = (
                valsdict[vname],
                tuple(bins_in_order),
            )
    return wrapped_up


def convert_correctionlib_file(filename):
    cset = correctionlib.CorrectionSet.from_file(filename)

    return {(key, "correctionlib_wrapper"): (cset[key],) for key in cset.keys()}



import awkward
import numpy
from copy import copy


def corrected_polar_met(met_pt, met_phi, jet_pt, jet_phi, jet_pt_orig, deltas=None):
    sj, cj = numpy.sin(jet_phi), numpy.cos(jet_phi)
    x = met_pt * numpy.cos(met_phi) + awkward.sum(
        jet_pt * cj - jet_pt_orig * cj, axis=1
    )
    y = met_pt * numpy.sin(met_phi) + awkward.sum(
        jet_pt * sj - jet_pt_orig * sj, axis=1
    )
    if deltas:
        positive, dx, dy = deltas
        x = x + dx if positive else x - dx
        y = y + dy if positive else y - dy
    return awkward.zip({"pt": numpy.hypot(x, y), "phi": numpy.arctan2(y, x)})


class CorrectedMETFactory(object):
    def __init__(self, name_map):
        for name in [
            "METpt",
            "METphi",
            "JetPt",
            "JetPhi",
            "ptRaw",
            "UnClusteredEnergyDeltaX",
            "UnClusteredEnergyDeltaY",
        ]:
            if name not in name_map or name_map[name] is None:
                raise ValueError(
                    f"There is no name mapping for {name}, which is needed for CorrectedMETFactory"
                )

        self.name_map = name_map

    def build(self, MET, corrected_jets, lazy_cache):
        if lazy_cache is None:
            raise Exception(
                "CorrectedMETFactory requires a awkward-array cache to function correctly."
            )
        lazy_cache = awkward._util.MappingProxy.maybe_wrap(lazy_cache)
        if not isinstance(MET, awkward.highlevel.Array) or not isinstance(
            corrected_jets, awkward.highlevel.Array
        ):
            raise Exception(
                "'MET' and 'corrected_jets' must be an awkward array of some kind!"
            )

        length = len(MET)
        form = awkward.forms.RecordForm(
            {
                "pt": MET[self.name_map["METpt"]].layout.form,
                "phi": MET[self.name_map["METphi"]].layout.form,
            },
        )

        def make_variant(*args):
            variant = copy(MET)
            corrected_met = awkward.virtual(
                corrected_polar_met,
                args=args,
                length=length,
                form=form,
                cache=lazy_cache,
            )
            variant[self.name_map["METpt"]] = awkward.virtual(
                lambda: awkward.materialized(corrected_met.pt),
                length=length,
                form=form.contents["pt"],
                cache=lazy_cache,
            )
            variant[self.name_map["METphi"]] = awkward.virtual(
                lambda: awkward.materialized(corrected_met.phi),
                length=length,
                form=form.contents["phi"],
                cache=lazy_cache,
            )
            return variant

        def lazy_variant(unc, metpt, metphi, jetpt, jetphi, jetptraw):
            return awkward.zip(
                {
                    "up": make_variant(
                        MET[metpt],
                        MET[metphi],
                        corrected_jets[unc].up[jetpt],
                        corrected_jets[unc].up[jetphi],
                        corrected_jets[unc].up[jetptraw],
                    ),
                    "down": make_variant(
                        MET[metpt],
                        MET[metphi],
                        corrected_jets[unc].down[jetpt],
                        corrected_jets[unc].down[jetphi],
                        corrected_jets[unc].down[jetptraw],
                    ),
                },
                depth_limit=1,
                with_name="METSystematic",
            )

        out = make_variant(
            MET[self.name_map["METpt"]],
            MET[self.name_map["METphi"]],
            corrected_jets[self.name_map["JetPt"]],
            corrected_jets[self.name_map["JetPhi"]],
            corrected_jets[self.name_map["ptRaw"]],
        )
        out[self.name_map["METpt"] + "_orig"] = MET[self.name_map["METpt"]]
        out[self.name_map["METphi"] + "_orig"] = MET[self.name_map["METphi"]]

        out_dict = {field: out[field] for field in awkward.fields(out)}

        out_dict["MET_UnclusteredEnergy"] = awkward.zip(
            {
                "up": make_variant(
                    MET[self.name_map["METpt"]],
                    MET[self.name_map["METphi"]],
                    corrected_jets[self.name_map["JetPt"]],
                    corrected_jets[self.name_map["JetPhi"]],
                    corrected_jets[self.name_map["ptRaw"]],
                    (
                        True,
                        MET[self.name_map["UnClusteredEnergyDeltaX"]],
                        MET[self.name_map["UnClusteredEnergyDeltaY"]],
                    ),
                ),
                "down": make_variant(
                    MET[self.name_map["METpt"]],
                    MET[self.name_map["METphi"]],
                    corrected_jets[self.name_map["JetPt"]],
                    corrected_jets[self.name_map["JetPhi"]],
                    corrected_jets[self.name_map["ptRaw"]],
                    (
                        False,
                        MET[self.name_map["UnClusteredEnergyDeltaX"]],
                        MET[self.name_map["UnClusteredEnergyDeltaY"]],
                    ),
                ),
            },
            depth_limit=1,
            with_name="METSystematic",
        )

        for unc in filter(
            lambda x: x.startswith(("JER", "JES")), awkward.fields(corrected_jets)
        ):
            out_dict[unc] = awkward.virtual(
                lazy_variant,
                args=(
                    unc,
                    self.name_map["METpt"],
                    self.name_map["METphi"],
                    self.name_map["JetPt"],
                    self.name_map["JetPhi"],
                    self.name_map["ptRaw"],
                ),
                length=length,
                cache={},
            )

        out_parms = out.layout.parameters
        out = awkward.zip(
            out_dict, depth_limit=1, parameters=out_parms, behavior=out.behavior
        )

        return out

    def uncertainties(self):
        return ["MET_UnclusteredEnergy"]



# Copyright (c) 2019, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import Callable, Dict, Optional
from dask.distributed import Client
from .executor import run_coffea_processor, Executor


class DaskExecutor(Executor):
    def __init__(
        self,
        client_addr: Optional[str] = None,
        provided_dask_client: Optional[Client] = None,
    ):
        """Create a Dask executor to process the analysis

        Args:
            client_addr (Optional[str]): If `None` then create a local cluster that runs
                                         in-process. Otherwise connect to an already
                                         existing cluster.
            provided_dask_client (Optional[Client]): Pass in an initialized Dask Client.
                                         This client must have asynchronous=True.
        """
        if not provided_dask_client:
            self.is_local = not client_addr

            self.dask = (
                Client(threads_per_worker=10, asynchronous=True)
                if self.is_local
                else Client(client_addr, asynchronous=True)
            )
        else:
            assert provided_dask_client.asynchronous
            self.dask = provided_dask_client

    def get_result_file_stream(self, datasource, title):
        if self.is_local:
            return datasource.stream_result_files(title)
        else:
            return datasource.stream_result_file_urls(title)

    def run_async_analysis(
        self,
        file_url: str,
        tree_name: Optional[str],
        data_type: str,
        meta_data: Dict[str, str],
        process_func: Callable,
    ):
        """Create a dask future for a dask task to run the analysis."""
        data_result = self.dask.submit(
            run_coffea_processor,
            events_url=file_url,
            tree_name=tree_name,
            data_type=data_type,
            meta_data=meta_data,
            proc=process_func,
        )

        return data_result



import awkward
import numpy
import warnings
from functools import partial
import operator


_stack_parts = ["jec", "junc", "jer", "jersf"]
_MIN_JET_ENERGY = numpy.array(1e-2, dtype=numpy.float32)
_ONE_F32 = numpy.array(1.0, dtype=numpy.float32)
_ZERO_F32 = numpy.array(0.0, dtype=numpy.float32)
_JERSF_FORM = {
    "class": "NumpyArray",
    "inner_shape": [3],
    "itemsize": 4,
    "format": "f",
    "primitive": "float32",
}


# we're gonna assume that the first record array we encounter is the flattened data
def rewrap_recordarray(layout, depth, data):
    if isinstance(layout, awkward.layout.RecordArray):
        return lambda: data
    return None


def awkward_rewrap(arr, like_what, gfunc):
    behavior = awkward._util.behaviorof(like_what)
    func = partial(gfunc, data=arr.layout)
    layout = awkward.operations.convert.to_layout(like_what)
    newlayout = awkward._util.recursively_apply(layout, func)
    return awkward._util.wrap(newlayout, behavior=behavior)


def rand_gauss(item, randomstate):
    def getfunction(layout, depth):
        if isinstance(layout, awkward.layout.NumpyArray) or not isinstance(
            layout, (awkward.layout.Content, awkward.partition.PartitionedArray)
        ):
            return lambda: awkward.layout.NumpyArray(
                randomstate.normal(size=len(layout)).astype(numpy.float32)
            )
        return None

    out = awkward._util.recursively_apply(
        awkward.operations.convert.to_layout(item), getfunction
    )
    assert out is not None
    return awkward._util.wrap(out, awkward._util.behaviorof(item))


def jer_smear(
    variation,
    forceStochastic,
    pt_gen,
    jetPt,
    etaJet,
    jet_energy_resolution,
    jet_resolution_rand_gauss,
    jet_energy_resolution_scale_factor,
):
    pt_gen = pt_gen if not forceStochastic else None

    if not isinstance(jetPt, awkward.highlevel.Array):
        raise Exception("'jetPt' must be an awkward array of some kind!")

    if forceStochastic:
        pt_gen = awkward.without_parameters(awkward.zeros_like(jetPt))

    jersmear = jet_energy_resolution * jet_resolution_rand_gauss
    jersf = jet_energy_resolution_scale_factor[:, variation]
    deltaPtRel = (jetPt - pt_gen) / jetPt
    doHybrid = (pt_gen > 0) & (numpy.abs(deltaPtRel) < 3 * jet_energy_resolution)

    detSmear = 1 + (jersf - 1) * deltaPtRel
    stochSmear = 1 + numpy.sqrt(numpy.maximum(jersf**2 - 1, 0)) * jersmear

    min_jet_pt = _MIN_JET_ENERGY / numpy.cosh(etaJet)
    min_jet_pt_corr = min_jet_pt / jetPt

    smearfact = awkward.where(doHybrid, detSmear, stochSmear)
    smearfact = awkward.where(
        (smearfact * jetPt) < min_jet_pt, min_jet_pt_corr, smearfact
    )

    def getfunction(layout, depth):
        if isinstance(layout, awkward.layout.NumpyArray) or not isinstance(
            layout, (awkward.layout.Content, awkward.partition.PartitionedArray)
        ):
            return lambda: awkward.layout.NumpyArray(smearfact)
        return None

    smearfact = awkward._util.recursively_apply(
        awkward.operations.convert.to_layout(jetPt), getfunction
    )
    smearfact = awkward._util.wrap(smearfact, awkward._util.behaviorof(jetPt))
    return smearfact


class CorrectedJetsFactory(object):
    def __init__(self, name_map, jec_stack):
        # from PhysicsTools/PatUtils/interface/SmearedJetProducerT.h#L283
        self.forceStochastic = False

        if "ptRaw" not in name_map or name_map["ptRaw"] is None:
            warnings.warn(
                "There is no name mapping for ptRaw,"
                " CorrectedJets will assume that <object>.pt is raw pt!"
            )
            name_map["ptRaw"] = name_map["JetPt"] + "_raw"
        self.treat_pt_as_raw = "ptRaw" not in name_map

        if "massRaw" not in name_map or name_map["massRaw"] is None:
            warnings.warn(
                "There is no name mapping for massRaw,"
                " CorrectedJets will assume that <object>.mass is raw pt!"
            )
            name_map["ptRaw"] = name_map["JetMass"] + "_raw"

        total_signature = set()
        for part in _stack_parts:
            attr = getattr(jec_stack, part)
            if attr is not None:
                total_signature.update(attr.signature)

        missing = total_signature - set(name_map.keys())
        if len(missing) > 0:
            raise Exception(
                f"Missing mapping of {missing} in name_map!"
                + " Cannot evaluate jet corrections!"
                + " Please supply mappings for these variables!"
            )

        if "ptGenJet" not in name_map:
            warnings.warn(
                'Input JaggedCandidateArray must have "ptGenJet" in order to apply hybrid JER smearing method. Stochastic smearing will be applied.'
            )
            self.forceStochastic = True

        self.real_sig = [v for k, v in name_map.items()]
        self.name_map = name_map
        self.jec_stack = jec_stack

    def uncertainties(self):
        out = ["JER"] if self.jec_stack.jer is not None else []
        if self.jec_stack.junc is not None:
            out.extend(["JES_{0}".format(unc) for unc in self.jec_stack.junc.levels])
        return out

    def build(self, jets, lazy_cache):
        if lazy_cache is None:
            raise Exception(
                "CorrectedJetsFactory requires a awkward-array cache to function correctly."
            )
        lazy_cache = awkward._util.MappingProxy.maybe_wrap(lazy_cache)
        if not isinstance(jets, awkward.highlevel.Array):
            raise Exception("'jets' must be an awkward > 1.0.0 array of some kind!")
        fields = awkward.fields(jets)
        if len(fields) == 0:
            raise Exception(
                "Empty record, please pass a jet object with at least {self.real_sig} defined!"
            )
        out = awkward.flatten(jets)
        wrap = partial(awkward_rewrap, like_what=jets, gfunc=rewrap_recordarray)
        scalar_form = awkward.without_parameters(
            out[self.name_map["ptRaw"]]
        ).layout.form

        in_dict = {field: out[field] for field in fields}
        out_dict = dict(in_dict)

        # take care of nominal JEC (no JER if available)
        out_dict[self.name_map["JetPt"] + "_orig"] = out_dict[self.name_map["JetPt"]]
        out_dict[self.name_map["JetMass"] + "_orig"] = out_dict[
            self.name_map["JetMass"]
        ]
        if self.treat_pt_as_raw:
            out_dict[self.name_map["ptRaw"]] = out_dict[self.name_map["JetPt"]]
            out_dict[self.name_map["massRaw"]] = out_dict[self.name_map["JetMass"]]

        jec_name_map = dict(self.name_map)
        jec_name_map["JetPt"] = jec_name_map["ptRaw"]
        jec_name_map["JetMass"] = jec_name_map["massRaw"]
        if self.jec_stack.jec is not None:
            jec_args = {
                k: out_dict[jec_name_map[k]] for k in self.jec_stack.jec.signature
            }
            out_dict["jet_energy_correction"] = self.jec_stack.jec.getCorrection(
                **jec_args, form=scalar_form, lazy_cache=lazy_cache
            )
        else:
            out_dict["jet_energy_correction"] = awkward.without_parameters(
                awkward.ones_like(out_dict[self.name_map["JetPt"]])
            )

        # finally the lazy binding to the JEC
        init_pt = partial(
            awkward.virtual,
            operator.mul,
            args=(out_dict["jet_energy_correction"], out_dict[self.name_map["ptRaw"]]),
            cache=lazy_cache,
        )
        init_mass = partial(
            awkward.virtual,
            operator.mul,
            args=(
                out_dict["jet_energy_correction"],
                out_dict[self.name_map["massRaw"]],
            ),
            cache=lazy_cache,
        )

        out_dict[self.name_map["JetPt"]] = init_pt(length=len(out), form=scalar_form)
        out_dict[self.name_map["JetMass"]] = init_mass(
            length=len(out), form=scalar_form
        )

        out_dict[self.name_map["JetPt"] + "_jec"] = out_dict[self.name_map["JetPt"]]
        out_dict[self.name_map["JetMass"] + "_jec"] = out_dict[self.name_map["JetMass"]]

        # in jer we need to have a stash for the intermediate JEC products
        has_jer = False
        if self.jec_stack.jer is not None and self.jec_stack.jersf is not None:
            has_jer = True
            jer_name_map = dict(self.name_map)
            jer_name_map["JetPt"] = jer_name_map["JetPt"] + "_jec"
            jer_name_map["JetMass"] = jer_name_map["JetMass"] + "_jec"

            jerargs = {
                k: out_dict[jer_name_map[k]] for k in self.jec_stack.jer.signature
            }
            out_dict["jet_energy_resolution"] = self.jec_stack.jer.getResolution(
                **jerargs, form=scalar_form, lazy_cache=lazy_cache
            )

            jersfargs = {
                k: out_dict[jer_name_map[k]] for k in self.jec_stack.jersf.signature
            }
            out_dict[
                "jet_energy_resolution_scale_factor"
            ] = self.jec_stack.jersf.getScaleFactor(
                **jersfargs, form=_JERSF_FORM, lazy_cache=lazy_cache
            )

            seeds = numpy.array(out_dict[self.name_map["JetPt"] + "_orig"])[
                [0, -1]
            ].view("i4")
            out_dict["jet_resolution_rand_gauss"] = awkward.virtual(
                rand_gauss,
                args=(
                    out_dict[self.name_map["JetPt"] + "_orig"],
                    numpy.random.Generator(numpy.random.PCG64(seeds)),
                ),
                cache=lazy_cache,
                length=len(out),
                form=scalar_form,
            )

            init_jerc = partial(
                awkward.virtual,
                jer_smear,
                args=(
                    0,
                    self.forceStochastic,
                    out_dict[jer_name_map["ptGenJet"]],
                    out_dict[jer_name_map["JetPt"]],
                    out_dict[jer_name_map["JetEta"]],
                    out_dict["jet_energy_resolution"],
                    out_dict["jet_resolution_rand_gauss"],
                    out_dict["jet_energy_resolution_scale_factor"],
                ),
                cache=lazy_cache,
            )
            out_dict["jet_energy_resolution_correction"] = init_jerc(
                length=len(out), form=scalar_form
            )

            init_pt_jer = partial(
                awkward.virtual,
                operator.mul,
                args=(
                    out_dict["jet_energy_resolution_correction"],
                    out_dict[jer_name_map["JetPt"]],
                ),
                cache=lazy_cache,
            )
            init_mass_jer = partial(
                awkward.virtual,
                operator.mul,
                args=(
                    out_dict["jet_energy_resolution_correction"],
                    out_dict[jer_name_map["JetMass"]],
                ),
                cache=lazy_cache,
            )
            out_dict[self.name_map["JetPt"]] = init_pt_jer(
                length=len(out), form=scalar_form
            )
            out_dict[self.name_map["JetMass"]] = init_mass_jer(
                length=len(out), form=scalar_form
            )

            out_dict[self.name_map["JetPt"] + "_jer"] = out_dict[self.name_map["JetPt"]]
            out_dict[self.name_map["JetMass"] + "_jer"] = out_dict[
                self.name_map["JetMass"]
            ]

            # JER systematics
            jerc_up = partial(
                awkward.virtual,
                jer_smear,
                args=(
                    1,
                    self.forceStochastic,
                    out_dict[jer_name_map["ptGenJet"]],
                    out_dict[jer_name_map["JetPt"]],
                    out_dict[jer_name_map["JetEta"]],
                    out_dict["jet_energy_resolution"],
                    out_dict["jet_resolution_rand_gauss"],
                    out_dict["jet_energy_resolution_scale_factor"],
                ),
                cache=lazy_cache,
            )
            up = awkward.flatten(jets)
            up["jet_energy_resolution_correction"] = jerc_up(
                length=len(out), form=scalar_form
            )
            init_pt_jer = partial(
                awkward.virtual,
                operator.mul,
                args=(
                    up["jet_energy_resolution_correction"],
                    out_dict[jer_name_map["JetPt"]],
                ),
                cache=lazy_cache,
            )
            init_mass_jer = partial(
                awkward.virtual,
                operator.mul,
                args=(
                    up["jet_energy_resolution_correction"],
                    out_dict[jer_name_map["JetMass"]],
                ),
                cache=lazy_cache,
            )
            up[self.name_map["JetPt"]] = init_pt_jer(length=len(out), form=scalar_form)
            up[self.name_map["JetMass"]] = init_mass_jer(
                length=len(out), form=scalar_form
            )

            jerc_down = partial(
                awkward.virtual,
                jer_smear,
                args=(
                    2,
                    self.forceStochastic,
                    out_dict[jer_name_map["ptGenJet"]],
                    out_dict[jer_name_map["JetPt"]],
                    out_dict[jer_name_map["JetEta"]],
                    out_dict["jet_energy_resolution"],
                    out_dict["jet_resolution_rand_gauss"],
                    out_dict["jet_energy_resolution_scale_factor"],
                ),
                cache=lazy_cache,
            )
            down = awkward.flatten(jets)
            down["jet_energy_resolution_correction"] = jerc_down(
                length=len(out), form=scalar_form
            )
            init_pt_jer = partial(
                awkward.virtual,
                operator.mul,
                args=(
                    down["jet_energy_resolution_correction"],
                    out_dict[jer_name_map["JetPt"]],
                ),
                cache=lazy_cache,
            )
            init_mass_jer = partial(
                awkward.virtual,
                operator.mul,
                args=(
                    down["jet_energy_resolution_correction"],
                    out_dict[jer_name_map["JetMass"]],
                ),
                cache=lazy_cache,
            )
            down[self.name_map["JetPt"]] = init_pt_jer(
                length=len(out), form=scalar_form
            )
            down[self.name_map["JetMass"]] = init_mass_jer(
                length=len(out), form=scalar_form
            )
            out_dict["JER"] = awkward.zip(
                {"up": up, "down": down}, depth_limit=1, with_name="JetSystematic"
            )

        if self.jec_stack.junc is not None:
            juncnames = {}
            juncnames.update(self.name_map)
            if has_jer:
                juncnames["JetPt"] = juncnames["JetPt"] + "_jer"
                juncnames["JetMass"] = juncnames["JetMass"] + "_jer"
            else:
                juncnames["JetPt"] = juncnames["JetPt"] + "_jec"
                juncnames["JetMass"] = juncnames["JetMass"] + "_jec"
            juncargs = {
                k: out_dict[juncnames[k]] for k in self.jec_stack.junc.signature
            }
            juncs = self.jec_stack.junc.getUncertainty(**juncargs)

            def junc_smeared_val(uncvals, up_down, variable):
                return awkward.materialized(uncvals[:, up_down] * variable)

            def build_variation(unc, jetpt, jetpt_orig, jetmass, jetmass_orig, updown):
                var_dict = dict(in_dict)
                var_dict[jetpt] = awkward.virtual(
                    junc_smeared_val,
                    args=(
                        unc,
                        updown,
                        jetpt_orig,
                    ),
                    length=len(out),
                    form=scalar_form,
                    cache=lazy_cache,
                )
                var_dict[jetmass] = awkward.virtual(
                    junc_smeared_val,
                    args=(
                        unc,
                        updown,
                        jetmass_orig,
                    ),
                    length=len(out),
                    form=scalar_form,
                    cache=lazy_cache,
                )
                return awkward.zip(
                    var_dict,
                    depth_limit=1,
                    parameters=out.layout.parameters,
                    behavior=out.behavior,
                )

            def build_variant(unc, jetpt, jetpt_orig, jetmass, jetmass_orig):
                up = build_variation(unc, jetpt, jetpt_orig, jetmass, jetmass_orig, 0)
                down = build_variation(unc, jetpt, jetpt_orig, jetmass, jetmass_orig, 1)
                return awkward.zip(
                    {"up": up, "down": down}, depth_limit=1, with_name="JetSystematic"
                )

            for name, func in juncs:
                out_dict[f"jet_energy_uncertainty_{name}"] = func
                out_dict[f"JES_{name}"] = build_variant(
                    func,
                    self.name_map["JetPt"],
                    out_dict[juncnames["JetPt"]],
                    self.name_map["JetMass"],
                    out_dict[juncnames["JetMass"]],
                )

        out_parms = out.layout.parameters
        out_parms["corrected"] = True
        out = awkward.zip(
            out_dict, depth_limit=1, parameters=out_parms, behavior=out.behavior
        )

        return wrap(out)



from coffea.jetmet_tools.FactorizedJetCorrector import FactorizedJetCorrector, _levelre
from coffea.jetmet_tools.JetResolution import JetResolution
from coffea.jetmet_tools.JetResolutionScaleFactor import JetResolutionScaleFactor
from coffea.jetmet_tools.JetCorrectionUncertainty import JetCorrectionUncertainty

_singletons = ["jer", "jersf"]
_nicenames = ["Jet Resolution Calculator", "Jet Resolution Scale Factor Calculator"]


class JECStack(object):
    def __init__(self, corrections, jec=None, junc=None, jer=None, jersf=None):
        """
        corrections is a dict-like of function names and functions
        we expect JEC names to be formatted as their filenames
        jecs, etc. can be overridden by passing in the appropriate corrector class.
        """
        self._jec = None
        self._junc = None
        self._jer = None
        self._jersf = None

        assembled = {"jec": {}, "junc": {}, "jer": {}, "jersf": {}}
        for key in corrections.keys():
            if "Uncertainty" in key:
                assembled["junc"][key] = corrections[key]
            elif "SF" in key:
                assembled["jersf"][key] = corrections[key]
            elif "Resolution" in key and "SF" not in key:
                assembled["jer"][key] = corrections[key]
            elif len(_levelre.findall(key)) > 0:
                assembled["jec"][key] = corrections[key]

        for corrtype, nname in zip(_singletons, _nicenames):
            Noftype = len(assembled[corrtype])
            if Noftype > 1:
                raise Exception(
                    f"JEC Stack has at most one {nname}, {Noftype} are present"
                )

        if jec is None:
            if len(assembled["jec"]) == 0:
                self._jec = None  # allow for no JEC
            else:
                self._jec = FactorizedJetCorrector(
                    **{name: corrections[name] for name in assembled["jec"]}
                )
        else:
            if isinstance(jec, FactorizedJetCorrector):
                self._jec = jec
            else:
                raise Exception(
                    'JECStack needs a FactorizedJetCorrector passed as "jec"'
                    + " got object of type {}".format(type(jec))
                )

        if junc is None:
            if len(assembled["junc"]) > 0:
                self._junc = JetCorrectionUncertainty(
                    **{name: corrections[name] for name in assembled["junc"]}
                )
        else:
            if isinstance(junc, JetCorrectionUncertainty):
                self._junc = junc
            else:
                raise Exception(
                    'JECStack needs a JetCorrectionUncertainty passed as "junc"'
                    + " got object of type {}".format(type(junc))
                )

        if jer is None:
            if len(assembled["jer"]) > 0:
                self._jer = JetResolution(
                    **{name: corrections[name] for name in assembled["jer"]}
                )
        else:
            if isinstance(jer, JetResolution):
                self._jer = jer
            else:
                raise Exception(
                    '"jer" must be of type "JetResolution"'
                    + " got {}".format(type(jer))
                )

        if jersf is None:
            if len(assembled["jersf"]) > 0:
                self._jersf = JetResolutionScaleFactor(
                    **{name: corrections[name] for name in assembled["jersf"]}
                )
        else:
            if isinstance(jer, JetResolutionScaleFactor):
                self._jersf = jersf
            else:
                raise Exception(
                    '"jer" must be of type "JetResolutionScaleFactor"'
                    + " got {}".format(type(jer))
                )

        if (self.jer is None) != (self.jersf is None):
            raise Exception("Cannot apply JER-SF without an input JER, and vice-versa!")

    @property
    def blank_name_map(self):
        out = {
            "massRaw",
            "ptRaw",
            "JetMass",
            "JetPt",
            "METpt",
            "METphi",
            "JetPhi",
            "UnClusteredEnergyDeltaX",
            "UnClusteredEnergyDeltaY",
        }
        if self._jec is not None:
            for name in self._jec.signature:
                out.add(name)
        if self._junc is not None:
            for name in self._junc.signature:
                out.add(name)
        if self._jer is not None:
            for name in self._jer.signature:
                out.add(name)
        if self._jersf is not None:
            for name in self._jersf.signature:
                out.add(name)
        return {name: None for name in out}

    @property
    def jec(self):
        return self._jec

    @property
    def junc(self):
        return self._junc

    @property
    def jer(self):
        return self._jer

    @property
    def jersf(self):
        return self._jersf



import urllib.parse


def quote(x):
    return urllib.parse.quote(x, safe="")


def unquote(x):
    return urllib.parse.unquote(x)


def concat(*keys):
    return quote(",".join(unquote(k) for k in keys))


def tuple_to_key(tup):
    return "/".join(quote(x) for x in tup)


def key_to_tuple(key):
    return tuple(unquote(x) for x in key.split("/"))



from ..util import numpy as np, deprecate
import uproot
import re

cycle = re.compile(r";\d+")


def killcycle(s, cycle):
    return cycle.sub("", s)


histTypes = ["TH1D", "TH1F", "TH2D", "TH2F", "TH3D", "TH3F"]
graphTypes = ["TGraphAsymmErrors", "TGraph2D"]


def convert_histo_root_file(file):
    converted_file = {}
    fin = uproot.open(file.strip())
    for path, item in fin.iteritems(recursive=True):
        if isinstance(item, uproot.ReadOnlyDirectory):
            continue
        nicepath = killcycle(path, cycle)
        rootclass = item.classname
        if rootclass in histTypes:
            edges = tuple(ax.edges() for ax in item.axes)
            converted_file[(nicepath, "dense_lookup")] = item.values(), edges
            if hasattr(item, "variances"):
                converted_file[(nicepath + "_error", "dense_lookup")] = (
                    np.sqrt(item.variances()),
                    edges,
                )
        elif rootclass in graphTypes:
            deprecate(
                RuntimeError(
                    "The support for TGraph-types will be removed in a future coffea release. In case you need support for TGraph-type, please submit an issue to `https://github.com/CoffeaTeam/coffea/issues`."  # noqa: E501
                ),
                "<unknown>",
            )
            continue
        elif hasattr(item, "_fEX") and hasattr(item, "_fEY"):
            # TODO what is this?
            tempArrX = item._fEX
            tempArrY = item._fEY
            converted_file[(nicepath, "dense_lookup")] = [tempArrX, tempArrY]

    return converted_file



from coffea.lookup_tools.dense_lookup import dense_lookup
from coffea.lookup_tools.dense_evaluated_lookup import dense_evaluated_lookup
from coffea.lookup_tools.jme_standard_function import jme_standard_function
from coffea.lookup_tools.jersf_lookup import jersf_lookup
from coffea.lookup_tools.jec_uncertainty_lookup import jec_uncertainty_lookup
from coffea.lookup_tools.rochester_lookup import rochester_lookup
from coffea.lookup_tools.correctionlib_wrapper import correctionlib_wrapper


lookup_types = {
    "dense_lookup": dense_lookup,
    "dense_evaluated_lookup": dense_evaluated_lookup,
    "jme_standard_function": jme_standard_function,
    "jersf_lookup": jersf_lookup,
    "jec_uncertainty_lookup": jec_uncertainty_lookup,
    "rochester_lookup": rochester_lookup,
    "correctionlib_wrapper": correctionlib_wrapper,
}


class evaluator(object):
    """
    The evaluator class serves as a single point of extry for
    looking up values of histograms and other functions read in
    with the extractor class. Stored look ups can be indexed by
    name and then called through an overloaded __call__ function.

    Example::

        #assuming 'eta' and 'pt' are array like objects
        wgts = "testSF2d scalefactors_Tight_Electron tests/samples/testSF2d.histo.root"
        extractor.add_weight_sets([wgts])
        extractor.finalize(reduce_list=['testSF2d'])
        evaluator = extractor.make_evaluator()
        out = evaluator["testSF2d"](eta, pt)

    The returned value has the same shape as the input arguments.

    lookup_types is a map of possible contructors for extracted data
    """

    def __init__(self, names, types, primitives):
        """
        initialize the evaluator from a list of inputs names,
        lookup types, and input primitive data
        """
        self._functions = {}
        for key in names.keys():
            lookup_type = types[names[key]]
            lookup_def = primitives[names[key]]
            self._functions[key] = lookup_types[lookup_type](*lookup_def)

    def __dir__(self):
        """dir is overloaded to list all available functions
        in the evaluator
        """
        return self._functions.keys()

    def __getitem__(self, key):
        """return a function named 'key'"""
        return self._functions[key]

    def keys(self):
        """returns the available functions"""
        return self._functions.keys()

    def __contains__(self, item):
        """item in X"""
        return item in self._functions



from __future__ import division
from collections import namedtuple
import numpy
from coffea.processor.accumulator import AccumulatorABC
import copy
import functools
import numbers
import re
import warnings
import awkward

# Python 2 and 3 compatibility
_regex_pattern = re.compile("dummy").__class__
try:
    basestring
except NameError:
    basestring = str

try:
    from collections.abc import Sequence
except ImportError:
    from collections import Sequence

MaybeSumSlice = namedtuple("MaybeSumSlice", ["start", "stop", "sum"])


def assemble_blocks(array, ndslice, depth=0):
    """
    Turns an n-dimensional slice of array (tuple of slices)
     into a nested list of numpy arrays that can be passed to numpy.block()

    Under the assumption that index 0 of any dimension is underflow, -2 overflow, -1 nanflow,
     this function will add the range not in the slice to the appropriate (over/under)flow bins
    """
    if depth == 0:
        ndslice = [MaybeSumSlice(s.start, s.stop, False) for s in ndslice]
    if depth == len(ndslice):
        slice_op = tuple(slice(s.start, s.stop) for s in ndslice)
        sum_op = tuple(i for i, s in enumerate(ndslice) if s.sum)
        return array[slice_op].sum(axis=sum_op, keepdims=True)
    slist = []
    newslice = ndslice[:]
    if ndslice[depth].start is not None:
        newslice[depth] = MaybeSumSlice(None, ndslice[depth].start, True)
        slist.append(assemble_blocks(array, newslice, depth + 1))
    newslice[depth] = MaybeSumSlice(ndslice[depth].start, ndslice[depth].stop, False)
    slist.append(assemble_blocks(array, newslice, depth + 1))
    if ndslice[depth].stop is not None:
        newslice[depth] = MaybeSumSlice(ndslice[depth].stop, -1, True)
        slist.append(assemble_blocks(array, newslice, depth + 1))
        newslice[depth] = MaybeSumSlice(-1, None, False)
        slist.append(assemble_blocks(array, newslice, depth + 1))
    return slist


def overflow_behavior(overflow):
    if overflow == "none":
        return slice(1, -2)
    elif overflow == "under":
        return slice(None, -2)
    elif overflow == "over":
        return slice(1, -1)
    elif overflow == "all":
        return slice(None, -1)
    elif overflow == "allnan":
        return slice(None)
    elif overflow == "justnan":
        return slice(-1, None)
    else:
        raise ValueError("Unrecognized overflow behavior: %s" % overflow)


@functools.total_ordering
class Interval(object):
    """Real number interval

    Totally ordered, assuming no overlap in intervals.
    A special nan interval can be constructed, which is defined
    as greater than ``[*, inf)``

    Parameters
    ----------
        lo : float
            Bin lower bound, inclusive
        hi : float
            Bin upper bound, exclusive
    """

    def __init__(self, lo, hi, label=None):
        self._lo = float(lo)
        self._hi = float(hi)
        self._label = label

    def __repr__(self):
        return "<%s (%s) instance at 0x%0x>" % (
            self.__class__.__name__,
            str(self),
            id(self),
        )

    def __str__(self):
        if self._label is not None:
            return self._label
        if self.nan():
            return "(nanflow)"
        # string representation of floats is apparently a touchy subject.. further reading:
        # https://stackoverflow.com/questions/25898733/why-does-strfloat-return-more-digits-in-python-3-than-python-2
        return "%s%.12g, %.12g)" % (
            "(" if self._lo == -numpy.inf else "[",
            self._lo,
            self._hi,
        )

    def __hash__(self):
        return hash((self._lo, self._hi))

    def __lt__(self, other):
        if other.nan() and not self.nan():
            return True
        elif self.nan():
            return False
        elif self._lo < other._lo:
            if self._hi > other._lo:
                raise ValueError(
                    "Intervals %r and %r intersect! What are you doing?!"
                    % (self, other)
                )
            return True
        return False

    def __eq__(self, other):
        if not isinstance(other, Interval):
            return False
        if other.nan() and self.nan():
            return True
        if self._lo == other._lo and self._hi == other._hi:
            return True
        return False

    def nan(self):
        return numpy.isnan(self._hi)

    @property
    def lo(self):
        """Lower boundary of this bin, inclusive"""
        return self._lo

    @property
    def hi(self):
        """Upper boundary of this bin, exclusive"""
        return self._hi

    @property
    def mid(self):
        """Midpoint of this bin"""
        return (self._hi + self._lo) / 2

    @property
    def label(self):
        """Label of this bin, mutable"""
        return self._label

    @label.setter
    def label(self, lbl):
        self._label = lbl


@functools.total_ordering
class StringBin(object):
    """A string used to fill a sparse axis

    Totally ordered, lexicographically by name.

    Parameters
    ----------
        name : str
            Name of the bin, as used in `Hist.fill` calls
        label : str
            The `str` representation of this bin can be overriden by
            a custom label, which will be used preferentially in legends
            produced by `hist.plot1d`, etc.
    """

    def __init__(self, name, label=None):
        if not isinstance(name, basestring):
            raise TypeError(
                "StringBin only supports string categories, received a %r" % name
            )
        elif "*" in name:
            raise ValueError(
                "StringBin does not support character '*' as it conflicts with wildcard mapping."
            )
        self._name = name
        self._label = label

    def __repr__(self):
        return "<%s (%s) instance at 0x%0x>" % (
            self.__class__.__name__,
            self.name,
            id(self),
        )

    def __str__(self):
        if self._label is not None:
            return self._label
        return self._name

    def __hash__(self):
        return hash(self._name)

    def __lt__(self, other):
        return self._name < other._name

    def __eq__(self, other):
        if isinstance(other, StringBin):
            return self._name == other._name
        return False

    @property
    def name(self):
        """Name of this bin, *Immutable*"""
        return self._name

    @property
    def label(self):
        """Label of this bin, mutable"""
        return self._label

    @label.setter
    def label(self, lbl):
        self._label = lbl


class Axis(object):
    """
    Axis: Base class for any type of axis
    Derived classes should implement, at least, an equality override
    """

    def __init__(self, name, label):
        if name == "weight":
            raise ValueError(
                "Cannot create axis: 'weight' is a reserved keyword for Hist.fill()"
            )
        self._name = name
        self._label = label

    def __repr__(self):
        return "<%s (name=%s) instance at 0x%0x>" % (
            self.__class__.__name__,
            self._name,
            id(self),
        )

    @property
    def name(self):
        return self._name

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, label):
        self._label = label

    def __eq__(self, other):
        if isinstance(other, Axis):
            if self._name != other._name:
                return False
            # label doesn't matter
            return True
        elif isinstance(other, basestring):
            # Convenient for testing axis in list by name
            if self._name != other:
                return False
            return True
        raise TypeError("Cannot compare an Axis with a %r" % other)


class SparseAxis(Axis):
    """
    SparseAxis: ABC for a sparse axis

    Derived should implement:
        **index(identifier)** - return a hashable object for indexing

        **__eq__(axis)** - axis has same definition (not necessarily same bins)

        **__getitem__(index)** - return an identifier

        **_ireduce(slice)** - return a list of hashes, slice is arbitrary

    What we really want here is a hashlist with some slice sugar on top
    It is usually the case that the identifier is already hashable,
    in which case index and __getitem__ are trivial, but this mechanism
    may be useful if the size of the tuple of identifiers in a
    sparse-binned histogram becomes too large
    """

    pass


class Cat(SparseAxis):
    """A category axis with name and label

    Parameters
    ----------
        name : str
            is used as a keyword in histogram filling, immutable
        label : str
            describes the meaning of the axis, can be changed
        sorting : {'identifier', 'placement', 'integral'}, optional
            Axis sorting when listing identifiers.  Default 'placement'
            Changing this setting can effect the order of stack plotting
            in `hist.plot1d`.

    The number of categories is arbitrary, and can be filled sparsely
    Identifiers are strings
    """

    def __init__(self, name, label, sorting="identifier"):
        super(Cat, self).__init__(name, label)
        # In all cases key == value.name
        self._bins = {}
        self._sorting = sorting
        self._sorted = []

    def index(self, identifier):
        """Index of a identifer or label

        Parameters
        ----------
            identifier : str or StringBin
                The identifier to lookup

        Returns a `StringBin` corresponding to the given argument (trival in the case
        where a `StringBin` was passed) and saves a reference internally in the case where
        the identifier was not seen before by this axis.
        """
        if isinstance(identifier, StringBin):
            index = identifier
        else:
            index = StringBin(identifier)
        if index.name not in self._bins:
            self._bins[index.name] = index
            self._sorted.append(index.name)
            if self._sorting == "identifier":
                self._sorted.sort()
        return self._bins[index.name]

    def __eq__(self, other):
        # Sparse, so as long as name is the same
        return super(Cat, self).__eq__(other)

    def __getitem__(self, index):
        if not isinstance(index, StringBin):
            raise TypeError("Expected a StringBin object, got: %r" % index)
        identifier = index.name
        if identifier not in self._bins:
            raise KeyError("No identifier %r in this Category axis")
        return identifier

    def _ireduce(self, the_slice):
        out = None
        if isinstance(the_slice, StringBin):
            out = [the_slice.name]
        elif isinstance(the_slice, _regex_pattern):
            out = [k for k in self._sorted if the_slice.match(k)]
        elif isinstance(the_slice, basestring):
            pattern = "^" + re.escape(the_slice).replace(r"\*", ".*") + "$"
            m = re.compile(pattern)
            out = [k for k in self._sorted if m.match(k)]
        elif isinstance(the_slice, list):
            if not all(k in self._sorted for k in the_slice):
                warnings.warn(
                    "Not all requested indices present in %r" % self, RuntimeWarning
                )
            out = [k for k in self._sorted if k in the_slice]
        elif isinstance(the_slice, slice):
            if the_slice.step is not None:
                raise IndexError("Not sure how to use slice step for categories...")
            start, stop = 0, len(self._sorted)
            if isinstance(the_slice.start, basestring):
                start = self._sorted.index(the_slice.start)
            else:
                start = the_slice.start
            if isinstance(the_slice.stop, basestring):
                stop = self._sorted.index(the_slice.stop)
            else:
                stop = the_slice.stop
            out = self._sorted[start:stop]
        else:
            raise IndexError(
                "Cannot understand slice %r on axis %r" % (the_slice, self)
            )
        return [self._bins[k] for k in out]

    @property
    def size(self):
        """Number of bins"""
        return len(self._bins)

    @property
    def sorting(self):
        """Sorting definition to adhere to

        See `Cat` constructor for possible values
        """
        return self._sorting

    @sorting.setter
    def sorting(self, newsorting):
        if newsorting == "placement":
            # not much we can do about already inserted values
            pass
        elif newsorting == "identifier":
            self._sorted.sort()
        elif newsorting == "integral":
            # this will be checked in any Hist.identifiers() call accessing this axis
            pass
        else:
            raise AttributeError("Invalid axis sorting type: %s" % newsorting)
        self._sorting = newsorting

    def identifiers(self):
        """List of `StringBin` identifiers"""
        return [self._bins[k] for k in self._sorted]


class DenseAxis(Axis):
    """
    DenseAxis: ABC for a fixed-size densely-indexed axis

    Derived should implement:
        **index(identifier)** - return an index

        **__eq__(axis)** - axis has same definition and binning

        **__getitem__(index)** - return an identifier

        **_ireduce(slice)** - return a slice or list of indices, input slice to be interpred as values

        **reduced(islice)** - return a new axis with binning corresponding to the index slice (from _ireduce)

    TODO: hasoverflow(), not all dense axes might have an overflow concept,
    currently it is implicitly assumed they do (as the only dense type is a numeric axis)
    """

    pass


class Bin(DenseAxis):
    """A binned axis with name, label, and binning.

    Parameters
    ----------
        name : str
            is used as a keyword in histogram filling, immutable
        label : str
            describes the meaning of the axis, can be changed
        n_or_arr : int or list or numpy.ndarray
            Integer number of bins, if uniform binning. Otherwise, a list or
            numpy 1D array of bin boundaries.
        lo : float, optional
            lower boundary of bin range, if uniform binning
        hi : float, optional
            upper boundary of bin range, if uniform binning

    This axis will generate frequencies for n+3 bins, special bin indices:
    ``0 = underflow, n+1 = overflow, n+2 = nanflow``
    Bin boundaries are [lo, hi)
    """

    def __init__(self, name, label, n_or_arr, lo=None, hi=None):
        super(Bin, self).__init__(name, label)
        self._lazy_intervals = None
        if isinstance(n_or_arr, (list, numpy.ndarray)):
            self._uniform = False
            self._bins = numpy.array(n_or_arr, dtype="d")
            if not all(numpy.sort(self._bins) == self._bins):
                raise ValueError("Binning not sorted!")
            self._lo = self._bins[0]
            self._hi = self._bins[-1]
            # to make searchsorted differentiate inf from nan
            self._bins = numpy.append(self._bins, numpy.inf)
            self._interval_bins = numpy.r_[-numpy.inf, self._bins, numpy.nan]
            self._bin_names = numpy.full(self._interval_bins[:-1].size, None)
        elif isinstance(n_or_arr, numbers.Integral):
            if lo is None or hi is None:
                raise TypeError(
                    "Interpreting n_or_arr as uniform binning, please specify lo and hi values"
                )
            self._uniform = True
            self._lo = lo
            self._hi = hi
            self._bins = n_or_arr
            self._interval_bins = numpy.r_[
                -numpy.inf,
                numpy.linspace(self._lo, self._hi, self._bins + 1),
                numpy.inf,
                numpy.nan,
            ]
            self._bin_names = numpy.full(self._interval_bins[:-1].size, None)
        else:
            raise TypeError(
                "Cannot understand n_or_arr (nbins or binning array) type %r" % n_or_arr
            )

    @property
    def _intervals(self):
        if not hasattr(self, "_lazy_intervals") or self._lazy_intervals is None:
            self._lazy_intervals = [
                Interval(low, high, bin)
                for low, high, bin in zip(
                    self._interval_bins[:-1], self._interval_bins[1:], self._bin_names
                )
            ]
        return self._lazy_intervals

    def __getstate__(self):
        if hasattr(self, "_lazy_intervals") and self._lazy_intervals is not None:
            self._bin_names = numpy.array(
                [interval.label for interval in self._lazy_intervals]
            )
        self.__dict__.pop("_lazy_intervals", None)
        return self.__dict__

    def __setstate__(self, d):
        if "_intervals" in d:  # convert old hists to new serialization format
            _old_intervals = d.pop("_intervals")
            interval_bins = [i._lo for i in _old_intervals] + [_old_intervals[-1]._hi]
            d["_interval_bins"] = numpy.array(interval_bins)
            d["_bin_names"] = numpy.array(
                [interval._label for interval in _old_intervals]
            )
        if "_interval_bins" in d and "_bin_names" not in d:
            d["_bin_names"] = numpy.full(d["_interval_bins"][:-1].size, None)
        self.__dict__ = d

    def index(self, identifier):
        """Index of a identifer or label

        Parameters
        ----------
            identifier : float or Interval or numpy.ndarray
                The identifier(s) to lookup.  Supports vectorized
                calls when a numpy 1D array of numbers is passed.

        Returns an integer corresponding to the index in the axis where the histogram would be filled.
        The integer range includes flow bins: ``0 = underflow, n+1 = overflow, n+2 = nanflow``
        """
        isarray = isinstance(identifier, (awkward.Array, numpy.ndarray))
        if isarray or isinstance(identifier, numbers.Number):
            if isarray:
                identifier = numpy.asarray(identifier)
            if self._uniform:
                idx = numpy.clip(
                    numpy.floor(
                        (identifier - self._lo)
                        * float(self._bins)
                        / (self._hi - self._lo)
                    )
                    + 1,
                    0,
                    self._bins + 1,
                )
                if isinstance(idx, numpy.ndarray):
                    idx[numpy.isnan(idx)] = self.size - 1
                    idx = idx.astype(int)
                elif numpy.isnan(idx):
                    idx = self.size - 1
                else:
                    idx = int(idx)
                return idx
            else:
                return numpy.searchsorted(self._bins, identifier, side="right")
        elif isinstance(identifier, Interval):
            if identifier.nan():
                return self.size - 1
            for idx, interval in enumerate(self._intervals):
                if interval._lo <= identifier._lo and interval._hi >= identifier._hi:
                    return idx
            raise ValueError(
                "Axis %r has no interval that fully contains identifier %r"
                % (self, identifier)
            )
        raise TypeError("Request bin indices with a identifier or 1-D array only")

    def __eq__(self, other):
        if isinstance(other, DenseAxis):
            if not super(Bin, self).__eq__(other):
                return False
            if self._uniform != other._uniform:
                return False
            if self._uniform and self._bins != other._bins:
                return False
            if not self._uniform and not all(self._bins == other._bins):
                return False
            return True
        return super(Bin, self).__eq__(other)

    def __getitem__(self, index):
        return self._intervals[index]

    def _ireduce(self, the_slice):
        if isinstance(the_slice, numbers.Number):
            the_slice = slice(the_slice, the_slice)
        elif isinstance(the_slice, Interval):
            if the_slice.nan():
                return slice(-1, None)
            lo = the_slice._lo if the_slice._lo > -numpy.inf else None
            hi = the_slice._hi if the_slice._hi < numpy.inf else None
            the_slice = slice(lo, hi)
        if isinstance(the_slice, slice):
            blo, bhi = None, None
            if the_slice.start is not None:
                if the_slice.start < self._lo:
                    raise ValueError(
                        "Reducing along axis %r: requested start %r exceeds bin boundaries (use open slicing, e.g. x[:stop])"
                        % (self, the_slice.start)
                    )
                if self._uniform:
                    blo_real = (the_slice.start - self._lo) * self._bins / (
                        self._hi - self._lo
                    ) + 1
                    blo = numpy.clip(
                        numpy.round(blo_real).astype(int), 0, self._bins + 1
                    )
                    if abs(blo - blo_real) > 1.0e-14:
                        warnings.warn(
                            "Reducing along axis %r: requested start %r between bin boundaries, no interpolation is performed"
                            % (self, the_slice.start),
                            RuntimeWarning,
                        )
                else:
                    if the_slice.start not in self._bins:
                        warnings.warn(
                            "Reducing along axis %r: requested start %r between bin boundaries, no interpolation is performed"
                            % (self, the_slice.start),
                            RuntimeWarning,
                        )
                    blo = self.index(the_slice.start)
            if the_slice.stop is not None:
                if the_slice.stop > self._hi:
                    raise ValueError(
                        "Reducing along axis %r: requested stop %r exceeds bin boundaries (use open slicing, e.g. x[start:])"
                        % (self, the_slice.stop)
                    )
                if self._uniform:
                    bhi_real = (the_slice.stop - self._lo) * self._bins / (
                        self._hi - self._lo
                    ) + 1
                    bhi = numpy.clip(
                        numpy.round(bhi_real).astype(int), 0, self._bins + 1
                    )
                    if abs(bhi - bhi_real) > 1.0e-14:
                        warnings.warn(
                            "Reducing along axis %r: requested stop %r between bin boundaries, no interpolation is performed"
                            % (self, the_slice.stop),
                            RuntimeWarning,
                        )
                else:
                    if the_slice.stop not in self._bins:
                        warnings.warn(
                            "Reducing along axis %r: requested stop %r between bin boundaries, no interpolation is performed"
                            % (self, the_slice.stop),
                            RuntimeWarning,
                        )
                    bhi = self.index(the_slice.stop)
                # Assume null ranges (start==stop) mean we want the bin containing the value
                if blo is not None and blo == bhi:
                    bhi += 1
            if the_slice.step is not None:
                raise NotImplementedError(
                    "Step slicing can be interpreted as a rebin factor"
                )
            return slice(blo, bhi, the_slice.step)
        elif isinstance(the_slice, list) and all(
            isinstance(v, Interval) for v in the_slice
        ):
            raise NotImplementedError("Slice histogram from list of intervals")
        raise IndexError("Cannot understand slice %r on axis %r" % (the_slice, self))

    def reduced(self, islice):
        """Return a new axis with reduced binning

        The new binning corresponds to the slice made on this axis.
        Overflow will be taken care of by ``Hist.__getitem__``

        Parameters
        ----------
            islice : slice
                ``islice.start`` and ``islice.stop`` should be None or within ``[1, ax.size() - 1]``
                This slice is usually as returned from ``Bin._ireduce``
        """
        if islice.step is not None:
            raise NotImplementedError(
                "Step slicing can be interpreted as a rebin factor"
            )
        if islice.start is None and islice.stop is None:
            return self
        if self._uniform:
            lo = self._lo
            ilo = 0
            if islice.start is not None:
                lo += (islice.start - 1) * (self._hi - self._lo) / self._bins
                ilo = islice.start - 1
            hi = self._hi
            ihi = self._bins
            if islice.stop is not None:
                hi = self._lo + (islice.stop - 1) * (self._hi - self._lo) / self._bins
                ihi = islice.stop - 1
            bins = ihi - ilo
            # TODO: remove this once satisfied it works
            rbins = (hi - lo) * self._bins / (self._hi - self._lo)
            assert abs(bins - rbins) < 1e-14, "%d %f %r" % (bins, rbins, self)
            ax = Bin(self._name, self._label, bins, lo, hi)
            return ax
        else:
            lo = None if islice.start is None else islice.start - 1
            hi = -1 if islice.stop is None else islice.stop
            bins = self._bins[slice(lo, hi)]
            ax = Bin(self._name, self._label, bins)
            return ax

    @property
    def size(self):
        """Number of bins, including overflow (i.e. ``n + 3``)"""
        if self._uniform:
            return self._bins + 3
        # (inf added at constructor)
        return len(self._bins) + 1

    def edges(self, overflow="none"):
        """Bin boundaries

        Parameters
        ----------
            overflow : str
                Create overflow and/or underflow bins by adding a bin of same width to each end.
                See `Hist.sum` description for the allowed values.
        """
        if self._uniform:
            out = numpy.linspace(self._lo, self._hi, self._bins + 1)
        else:
            out = self._bins[:-1].copy()
        out = numpy.r_[
            2 * out[0] - out[1], out, 2 * out[-1] - out[-2], 3 * out[-1] - 2 * out[-2]
        ]
        return out[overflow_behavior(overflow)]

    def centers(self, overflow="none"):
        """Bin centers

        Parameters
        ----------
            overflow : str
                Create overflow and/or underflow bins by adding a bin of same width to each end.
                See `Hist.sum` description for the allowed values.
        """
        edges = self.edges(overflow)
        return (edges[:-1] + edges[1:]) / 2

    def identifiers(self, overflow="none"):
        """List of `Interval` identifiers"""
        return self._intervals[overflow_behavior(overflow)]


class Hist(AccumulatorABC):
    """
    Specify a multidimensional histogram.

    Parameters
    ----------
        label : str
            A description of the meaning of the sum of weights
        ``*axes``
            positional list of `Cat` or `Bin` objects, denoting the axes of the histogram
        axes : collections.abc.Sequence
            list of `Cat` or `Bin` objects, denoting the axes of the histogram (overridden by ``*axes``)
        dtype : str
            Underlying numpy dtype to use for storing sum of weights

    Examples
    --------

    Creating a histogram with a sparse axis, and two dense axes::

        h = coffea.hist.Hist("Observed bird count",
                             coffea.hist.Cat("species", "Bird species"),
                             coffea.hist.Bin("x", "x coordinate [m]", 20, -5, 5),
                             coffea.hist.Bin("y", "y coordinate [m]", 20, -5, 5),
                             )

        # or

        h = coffea.hist.Hist(label="Observed bird count",
                             axes=(coffea.hist.Cat("species", "Bird species"),
                                   coffea.hist.Bin("x", "x coordinate [m]", 20, -5, 5),
                                   coffea.hist.Bin("y", "y coordinate [m]", 20, -5, 5),
                                  )
                             )

        # or

        h = coffea.hist.Hist(axes=[coffea.hist.Cat("species", "Bird species"),
                                   coffea.hist.Bin("x", "x coordinate [m]", 20, -5, 5),
                                   coffea.hist.Bin("y", "y coordinate [m]", 20, -5, 5),
                                  ],
                             label="Observed bird count",
                             )

    which produces:

    >>> h
    <Hist (species,x,y) instance at 0x10d84b550>

    """

    #: Default numpy dtype to store sum of weights
    DEFAULT_DTYPE = "d"

    def __init__(self, label, *axes, **kwargs):
        if not isinstance(label, basestring):
            raise TypeError("label must be a string")
        self._label = label
        self._dtype = kwargs.pop(
            "dtype", Hist.DEFAULT_DTYPE
        )  # Much nicer in python3 :(
        self._axes = axes
        if len(axes) == 0 and "axes" in kwargs:
            if not isinstance(kwargs["axes"], Sequence):
                raise TypeError("axes must be a sequence type! (tuple, list, etc.)")
            self._axes = tuple(kwargs["axes"])
        elif len(axes) != 0 and "axes" in kwargs:
            warnings.warn(
                "axes defined by both positional arguments and keyword argument, using positional arguments"
            )

        if not all(isinstance(ax, Axis) for ax in self._axes):
            del self._axes
            raise TypeError("All axes must be derived from Axis class")
        # if we stably partition axes to sparse, then dense, some things simplify
        # ..but then the user would then see the order change under them
        self._dense_shape = tuple(
            [ax.size for ax in self._axes if isinstance(ax, DenseAxis)]
        )
        if numpy.prod(self._dense_shape) > 10000000:
            warnings.warn("Allocating a large (>10M bin) histogram!", RuntimeWarning)
        self._sumw = {}
        # Storage of sumw2 starts at first use of weight keyword in fill()
        self._sumw2 = None

    def __repr__(self):
        return "<%s (%s) instance at 0x%0x>" % (
            self.__class__.__name__,
            ",".join(d.name for d in self.axes()),
            id(self),
        )

    @property
    def label(self):
        """A label describing the meaning of the sum of weights"""
        return self._label

    @label.setter
    def label(self, label):
        self._label = label

    def copy(self, content=True):
        """Create a deep copy

        Parameters
        ----------
            content : bool
                If set false, only the histogram definition is copied, resetting
                the sum of weights to zero
        """
        out = Hist(self._label, *self._axes, dtype=self._dtype)
        if self._sumw2 is not None:
            out._sumw2 = {}
        if content:
            out._sumw = copy.deepcopy(self._sumw)
            out._sumw2 = copy.deepcopy(self._sumw2)
        return out

    def identity(self):
        """The identity (zero value) of this accumulator"""
        return self.copy(content=False)

    def clear(self):
        """Clear all content in this histogram"""
        self._sumw = {}
        self._sumw2 = None

    def axis(self, axis_name):
        """Get an ``Axis`` object"""
        if axis_name in self._axes:
            return self._axes[self._axes.index(axis_name)]
        raise KeyError("No axis %s found in %r" % (axis_name, self))

    def axes(self):
        """Get all axes in this histogram"""
        return self._axes

    @property
    def fields(self):
        """This is a stub for histbook compatibility"""
        return [ax.name for ax in self._axes]

    def dim(self):
        """Dimension of this histogram (number of axes)"""
        return len(self._axes)

    def dense_dim(self):
        """Dense dimension of this histogram (number of non-sparse axes)"""
        return len(self._dense_shape)

    def sparse_dim(self):
        """Sparse dimension of this histogram (number of sparse axes)"""
        return self.dim() - self.dense_dim()

    def dense_axes(self):
        """All dense axes"""
        return [ax for ax in self._axes if isinstance(ax, DenseAxis)]

    def sparse_axes(self):
        """All sparse axes"""
        return [ax for ax in self._axes if isinstance(ax, SparseAxis)]

    def sparse_nbins(self):
        """Total number of sparse bins"""
        return len(self._sumw)

    def _idense(self, axis):
        return self.dense_axes().index(axis)

    def _isparse(self, axis):
        return self.sparse_axes().index(axis)

    def _init_sumw2(self):
        self._sumw2 = {}
        for key in self._sumw.keys():
            self._sumw2[key] = self._sumw[key].copy()

    def compatible(self, other):
        """Checks if this histogram is compatible with another, i.e. they have identical binning"""
        if self.dim() != other.dim():
            return False
        if set(d.name for d in self.sparse_axes()) != set(
            d.name for d in other.sparse_axes()
        ):
            return False
        if not all(d1 == d2 for d1, d2 in zip(self.dense_axes(), other.dense_axes())):
            return False
        return True

    def add(self, other):
        """Add another histogram into this one, in-place"""
        if not self.compatible(other):
            raise ValueError(
                "Cannot add this histogram with histogram %r of dissimilar dimensions"
                % other
            )

        raxes = other.sparse_axes()

        def add_dict(left, right):
            for rkey in right.keys():
                lkey = tuple(
                    self.axis(rax).index(rax[ridx]) for rax, ridx in zip(raxes, rkey)
                )
                if lkey in left:
                    left[lkey] += right[rkey]
                else:
                    left[lkey] = copy.deepcopy(right[rkey])

        if self._sumw2 is None and other._sumw2 is None:
            pass
        elif self._sumw2 is None:
            self._init_sumw2()
            add_dict(self._sumw2, other._sumw2)
        elif other._sumw2 is None:
            add_dict(self._sumw2, other._sumw)
        else:
            add_dict(self._sumw2, other._sumw2)
        add_dict(self._sumw, other._sumw)
        return self

    def __getitem__(self, keys):
        if not isinstance(keys, tuple):
            keys = (keys,)
        if len(keys) > self.dim():
            raise IndexError("Too many indices for this histogram")
        elif len(keys) < self.dim():
            if Ellipsis in keys:
                idx = keys.index(Ellipsis)
                slices = (slice(None),) * (self.dim() - len(keys) + 1)
                keys = keys[:idx] + slices + keys[idx + 1 :]
            else:
                slices = (slice(None),) * (self.dim() - len(keys))
                keys += slices
        sparse_idx = []
        dense_idx = []
        new_dims = []
        for s, ax in zip(keys, self._axes):
            if isinstance(ax, SparseAxis):
                sparse_idx.append(ax._ireduce(s))
                new_dims.append(ax)
            else:
                islice = ax._ireduce(s)
                dense_idx.append(islice)
                new_dims.append(ax.reduced(islice))
        dense_idx = tuple(dense_idx)

        def dense_op(array):
            return numpy.block(assemble_blocks(array, dense_idx))

        out = Hist(self._label, *new_dims, dtype=self._dtype)
        if self._sumw2 is not None:
            out._init_sumw2()
        for sparse_key in self._sumw:
            if not all(k in idx for k, idx in zip(sparse_key, sparse_idx)):
                continue
            if sparse_key in out._sumw:
                out._sumw[sparse_key] += dense_op(self._sumw[sparse_key])
                if self._sumw2 is not None:
                    out._sumw2[sparse_key] += dense_op(self._sumw2[sparse_key])
            else:
                out._sumw[sparse_key] = dense_op(self._sumw[sparse_key]).copy()
                if self._sumw2 is not None:
                    out._sumw2[sparse_key] = dense_op(self._sumw2[sparse_key]).copy()
        return out

    def fill(self, **values):
        """Fill sum of weights from columns

        Parameters
        ----------
            ``**values``
                Keyword arguments, one for each axis name, of either flat numpy arrays
                (for dense dimensions) or literals (for sparse dimensions) which will
                be used to fill bins at the corresponding indices.

        Note
        ----
            The reserved keyword ``weight``, if specified, will increment sum of weights
            by the given column values, which must be broadcastable to the same dimension as all other
            columns.  Upon first use, this will trigger the storage of the sum of squared weights.


        Examples
        --------

        Filling the histogram from the `Hist` example:

        >>> h.fill(species='ducks', x=numpy.random.normal(size=10), y=numpy.random.normal(size=10), weight=numpy.ones(size=10) * 3)

        """
        weight = values.pop("weight", None)
        if isinstance(weight, (awkward.Array, numpy.ndarray)):
            weight = numpy.asarray(weight)
        if isinstance(weight, numbers.Number):
            weight = numpy.atleast_1d(weight)
        if not all(d.name in values for d in self._axes):
            missing = ", ".join(d.name for d in self._axes if d.name not in values)
            raise ValueError(
                "Not all axes specified for %r.  Missing: %s" % (self, missing)
            )
        if not all(name in self._axes for name in values):
            extra = ", ".join(name for name in values if name not in self._axes)
            raise ValueError(
                "Unrecognized axes specified for %r.  Extraneous: %s" % (self, extra)
            )

        if weight is not None and self._sumw2 is None:
            self._init_sumw2()

        sparse_key = tuple(d.index(values[d.name]) for d in self.sparse_axes())
        if sparse_key not in self._sumw:
            self._sumw[sparse_key] = numpy.zeros(
                shape=self._dense_shape, dtype=self._dtype
            )
            if self._sumw2 is not None:
                self._sumw2[sparse_key] = numpy.zeros(
                    shape=self._dense_shape, dtype=self._dtype
                )

        if self.dense_dim() > 0:
            dense_indices = tuple(
                d.index(values[d.name]) for d in self._axes if isinstance(d, DenseAxis)
            )
            xy = numpy.atleast_1d(
                numpy.ravel_multi_index(dense_indices, self._dense_shape)
            )
            if weight is not None:
                self._sumw[sparse_key][:] += numpy.bincount(
                    xy, weights=weight, minlength=numpy.array(self._dense_shape).prod()
                ).reshape(self._dense_shape)
                self._sumw2[sparse_key][:] += numpy.bincount(
                    xy,
                    weights=weight**2,
                    minlength=numpy.array(self._dense_shape).prod(),
                ).reshape(self._dense_shape)
            else:
                self._sumw[sparse_key][:] += numpy.bincount(
                    xy, weights=None, minlength=numpy.array(self._dense_shape).prod()
                ).reshape(self._dense_shape)
                if self._sumw2 is not None:
                    self._sumw2[sparse_key][:] += numpy.bincount(
                        xy,
                        weights=None,
                        minlength=numpy.array(self._dense_shape).prod(),
                    ).reshape(self._dense_shape)
        else:
            if weight is not None:
                self._sumw[sparse_key] += numpy.sum(weight)
                self._sumw2[sparse_key] += numpy.sum(weight**2)
            else:
                self._sumw[sparse_key] += 1.0
                if self._sumw2 is not None:
                    self._sumw2[sparse_key] += 1.0

    def sum(self, *axes, **kwargs):
        """Integrates out a set of axes, producing a new histogram

        Parameters
        ----------
            ``*axes``
                Positional list of axes to integrate out (either a string or an Axis object)

            overflow : {'none', 'under', 'over', 'all', 'allnan'}, optional
                How to treat the overflow bins in the sum.  Only applies to dense axes.
                'all' includes both under- and over-flow but not nan-flow bins.
                Default is 'none'.
        """
        overflow = kwargs.pop("overflow", "none")
        axes = [self.axis(ax) for ax in axes]
        reduced_dims = [ax for ax in self._axes if ax not in axes]
        out = Hist(self._label, *reduced_dims, dtype=self._dtype)
        if self._sumw2 is not None:
            out._init_sumw2()

        sparse_drop = []
        dense_slice = [slice(None)] * self.dense_dim()
        dense_sum_dim = []
        for axis in axes:
            if isinstance(axis, DenseAxis):
                idense = self._idense(axis)
                dense_sum_dim.append(idense)
                dense_slice[idense] = overflow_behavior(overflow)
            elif isinstance(axis, SparseAxis):
                isparse = self._isparse(axis)
                sparse_drop.append(isparse)
        dense_slice = tuple(dense_slice)
        dense_sum_dim = tuple(dense_sum_dim)

        def dense_op(array):
            if len(dense_sum_dim) > 0:
                return numpy.sum(array[dense_slice], axis=dense_sum_dim)
            return array

        for key in self._sumw.keys():
            new_key = tuple(k for i, k in enumerate(key) if i not in sparse_drop)
            if new_key in out._sumw:
                out._sumw[new_key] += dense_op(self._sumw[key])
                if self._sumw2 is not None:
                    out._sumw2[new_key] += dense_op(self._sumw2[key])
            else:
                out._sumw[new_key] = dense_op(self._sumw[key]).copy()
                if self._sumw2 is not None:
                    out._sumw2[new_key] = dense_op(self._sumw2[key]).copy()
        return out

    def project(self, *axes, **kwargs):
        """Project histogram onto a subset of its axes

        Parameters
        ----------
            ``*axes`` : str or Axis
                Positional list of axes to project on to
            overflow : str
                Controls behavior of integration over remaining axes.
                See `sum` description for meaning of allowed values
                Default is to *not include* overflow bins
        """
        overflow = kwargs.pop("overflow", "none")
        axes = [self.axis(ax) for ax in axes]
        toremove = [ax for ax in self.axes() if ax not in axes]
        return self.sum(*toremove, overflow=overflow)

    def integrate(self, axis_name, int_range=slice(None), overflow="none"):
        """Integrates current histogram along one dimension

        Parameters
        ----------
            axis_name : str or Axis
                Which dimension to reduce on
            int_range : slice
                Any slice, list, string, or other object that the axis will understand
                Default is to integrate over the whole range
            overflow : str
                See `sum` description for meaning of allowed values
                Default is to *not include* overflow bins

        """
        axis = self.axis(axis_name)
        full_slice = tuple(
            slice(None) if ax != axis else int_range for ax in self._axes
        )
        if isinstance(int_range, Interval):
            # Handle overflow intervals nicely
            if int_range.nan():
                overflow = "justnan"
            elif int_range.lo == -numpy.inf:
                overflow = "under"
            elif int_range.hi == numpy.inf:
                overflow = "over"
        return self[full_slice].sum(
            axis.name, overflow=overflow
        )  # slice may make new axis, use name

    def remove(self, bins, axis):
        """Remove bins from a sparse axis

        Parameters
        ----------
            bins : iterable
                A list of bin identifiers to remove
            axis : str or Axis
                Axis name or SparseAxis instance

        Returns a *copy* of the histogram with specified bins removed, not an in-place operation
        """
        axis = self.axis(axis)
        if not isinstance(axis, SparseAxis):
            raise NotImplementedError(
                "Hist.remove() only supports removing items from a sparse axis."
            )
        bins = [axis.index(binid) for binid in bins]
        keep = [binid.name for binid in self.identifiers(axis) if binid not in bins]
        full_slice = tuple(slice(None) if ax != axis else keep for ax in self._axes)
        return self[full_slice]

    def group(self, old_axes, new_axis, mapping, overflow="none"):
        """Group a set of slices on old axes into a single new axis

        Parameters
        ----------
            old_axes
                Axis or tuple of axes which are being grouped
            new_axis
                A new sparse dimension definition, e.g. a `Cat` instance
            mapping : dict
                A mapping ``{'new_bin': (slice, ...), ...}`` where each
                slice is on the axes being re-binned.  In the case of
                a single axis for ``old_axes``, ``{'new_bin': slice, ...}``
                is admissible.
            overflow : str
                See `sum` description for meaning of allowed values
                Default is to *not include* overflow bins

        Returns a new histogram object
        """
        if not isinstance(new_axis, SparseAxis):
            raise TypeError(
                "New axis must be a sparse axis.  Note: Hist.group() signature has changed to group(old_axes, new_axis, ...)!"
            )
        if new_axis in self.axes() and self.axis(new_axis) is new_axis:
            raise RuntimeError(
                "new_axis is already in the list of axes.  Note: Hist.group() signature has changed to group(old_axes, new_axis, ...)!"
            )
        if not isinstance(old_axes, tuple):
            old_axes = (old_axes,)
        old_axes = [self.axis(ax) for ax in old_axes]
        old_indices = [i for i, ax in enumerate(self._axes) if ax in old_axes]
        new_dims = [new_axis] + [ax for ax in self._axes if ax not in old_axes]
        out = Hist(self._label, *new_dims, dtype=self._dtype)
        if self._sumw2 is not None:
            out._init_sumw2()
        for new_cat in mapping.keys():
            the_slice = mapping[new_cat]
            if not isinstance(the_slice, tuple):
                the_slice = (the_slice,)
            if len(the_slice) != len(old_axes):
                raise Exception("Slicing does not match number of axes being rebinned")
            full_slice = [slice(None)] * self.dim()
            for idx, s in zip(old_indices, the_slice):
                full_slice[idx] = s
            full_slice = tuple(full_slice)
            reduced_hist = self[full_slice].sum(
                *tuple(ax.name for ax in old_axes), overflow=overflow
            )  # slice may change old axis binning
            new_idx = new_axis.index(new_cat)
            for key in reduced_hist._sumw:
                new_key = (new_idx,) + key
                out._sumw[new_key] = reduced_hist._sumw[key]
                if self._sumw2 is not None:
                    out._sumw2[new_key] = reduced_hist._sumw2[key]
        return out

    def rebin(self, old_axis, new_axis):
        """Rebin a dense axis

        This function will construct the mapping from old to new axis, and
        constructs a new histogram, rebinning the sum of weights along that dimension.

        Note
        ----
        No interpolation is performed, so the user must be sure the old
        and new axes have compatible bin boundaries, e.g. that they evenly
        divide each other.

        Parameters
        ----------
            old_axis : str or Axis
                Axis to rebin
            new_axis : str or Axis or int
                A DenseAxis object defining the new axis (e.g. a `Bin` instance).
                If a number N is supplied, the old axis edges are downsampled by N,
                resulting in a histogram with ``old_nbins // N`` bins.

        Returns a new `Hist` object.
        """
        old_axis = self.axis(old_axis)
        if isinstance(new_axis, numbers.Integral):
            new_axis = Bin(old_axis.name, old_axis.label, old_axis.edges()[::new_axis])
        new_dims = [ax if ax != old_axis else new_axis for ax in self._axes]
        out = Hist(self._label, *new_dims, dtype=self._dtype)
        if self._sumw2 is not None:
            out._init_sumw2()

        # would have been nice to use ufunc.reduceat, but we should support arbitrary reshuffling
        idense = self._idense(old_axis)

        def view_ax(idx):
            fullindex = [slice(None)] * self.dense_dim()
            fullindex[idense] = idx
            return tuple(fullindex)

        binmap = [new_axis.index(i) for i in old_axis.identifiers(overflow="allnan")]

        def dense_op(array):
            anew = numpy.zeros(out._dense_shape, dtype=out._dtype)
            for iold, inew in enumerate(binmap):
                anew[view_ax(inew)] += array[view_ax(iold)]
            return anew

        for key in self._sumw:
            out._sumw[key] = dense_op(self._sumw[key])
            if self._sumw2 is not None:
                out._sumw2[key] = dense_op(self._sumw2[key])
        return out

    def values(self, sumw2=False, overflow="none"):
        """Extract the sum of weights arrays from this histogram

        Parameters
        ----------
            sumw2 : bool
                If True, frequencies is a tuple of arrays (sum weights, sum squared weights)
            overflow
                See `sum` description for meaning of allowed values

        Returns a mapping ``{(sparse identifier, ...): numpy.array(...), ...}``
        where each array has dimension `dense_dim` and shape matching
        the number of bins per axis, plus 0-3 overflow bins depending
        on the ``overflow`` argument.
        """

        def view_dim(arr):
            if self.dense_dim() == 0:
                return arr
            else:
                return arr[
                    tuple(overflow_behavior(overflow) for _ in range(self.dense_dim()))
                ]

        out = {}
        for sparse_key in self._sumw.keys():
            id_key = tuple(ax[k] for ax, k in zip(self.sparse_axes(), sparse_key))
            if sumw2:
                if self._sumw2 is not None:
                    w2 = view_dim(self._sumw2[sparse_key])
                else:
                    w2 = view_dim(self._sumw[sparse_key])
                out[id_key] = (view_dim(self._sumw[sparse_key]), w2)
            else:
                out[id_key] = view_dim(self._sumw[sparse_key])
        return out

    def scale(self, factor, axis=None):
        """Scale histogram in-place by factor

        Parameters
        ----------
            factor : float or dict
                A number or mapping of identifier to number
            axis : optional
                Which (sparse) axis the dict applies to, may be a tuples of axes.
                The dict keys must follow the same structure.

        Examples
        --------
        This function is useful to quickly reweight according to some
        weight mapping along a sparse axis, such as the ``species`` axis
        in the `Hist` example:

        >>> h.scale({'ducks': 0.3, 'geese': 1.2}, axis='species')
        >>> h.scale({('ducks',): 0.5}, axis=('species',))
        >>> h.scale({('geese', 'honk'): 5.0}, axis=('species', 'vocalization'))
        """
        if self._sumw2 is None:
            self._init_sumw2()
        if isinstance(factor, numbers.Number) and axis is None:
            for key in self._sumw.keys():
                self._sumw[key] *= factor
                self._sumw2[key] *= factor**2
        elif isinstance(factor, dict):
            if not isinstance(axis, tuple):
                axis = (axis,)
                factor = {(k,): v for k, v in factor.items()}
            axis = tuple(map(self.axis, axis))
            isparse = list(map(self._isparse, axis))
            factor = {
                tuple(a.index(e) for a, e in zip(axis, k)): v for k, v in factor.items()
            }
            for key in self._sumw.keys():
                factor_key = tuple(key[i] for i in isparse)
                if factor_key in factor:
                    self._sumw[key] *= factor[factor_key]
                    self._sumw2[key] *= factor[factor_key] ** 2
        elif isinstance(factor, numpy.ndarray):
            axis = self.axis(axis)
            raise NotImplementedError("Scale dense dimension by a factor")
        else:
            raise TypeError("Could not interpret scale factor")

    def identifiers(self, axis, overflow="none"):
        """Return a list of identifiers for an axis

        Parameters
        ----------
            axis
                Axis name or Axis object
            overflow
                See `sum` description for meaning of allowed values
        """
        axis = self.axis(axis)
        if isinstance(axis, SparseAxis):
            out = []
            isparse = self._isparse(axis)
            for identifier in axis.identifiers():
                if any(k[isparse] == axis.index(identifier) for k in self._sumw.keys()):
                    out.append(identifier)
            if axis.sorting == "integral":
                hproj = {
                    key[0]: integral
                    for key, integral in self.project(axis).values().items()
                }
                out.sort(key=lambda k: hproj[k.name])
            return out
        elif isinstance(axis, DenseAxis):
            return axis.identifiers(overflow=overflow)

    def to_boost(self):
        """Convert this coffea Hist object to a boost_histogram obbject"""
        import boost_histogram

        newaxes = []
        for axis in self.axes():
            if isinstance(axis, Bin) and axis._uniform:
                newaxis = boost_histogram.axis.Regular(
                    axis._bins,
                    axis._lo,
                    axis._hi,
                    underflow=True,
                    overflow=True,
                )
                newaxis.name = axis.name
                newaxis.label = axis.label
                newaxes.append(newaxis)
            elif isinstance(axis, Bin) and not axis._uniform:
                newaxis = boost_histogram.axis.Variable(
                    axis.edges(),
                    underflow=True,
                    overflow=True,
                )
                newaxis.name = axis.name
                newaxis.label = axis.label
                newaxes.append(newaxis)
            elif isinstance(axis, Cat):
                identifiers = self.identifiers(axis)
                newaxis = boost_histogram.axis.StrCategory(
                    [x.name for x in identifiers],
                    growth=True,
                )
                newaxis.name = axis.name
                newaxis.label = axis.label
                newaxis.bin_labels = [x.label for x in identifiers]
                newaxes.append(newaxis)

        if self._sumw2 is None:
            storage = boost_histogram.storage.Double()
        else:
            storage = boost_histogram.storage.Weight()

        out = boost_histogram.Histogram(*newaxes, storage=storage)
        out.label = self.label

        def expandkey(key):
            kiter = iter(key)
            for ax in newaxes:
                if isinstance(ax, boost_histogram.axis.StrCategory):
                    yield ax.index(next(kiter))
                else:
                    yield slice(None)

        if self._sumw2 is None:
            values = self.values(overflow="all")
            for sparse_key, sumw in values.items():
                index = tuple(expandkey(sparse_key))
                view = out.view(flow=True)
                view[index] = sumw
        else:
            values = self.values(sumw2=True, overflow="all")
            for sparse_key, (sumw, sumw2) in values.items():
                index = tuple(expandkey(sparse_key))
                view = out.view(flow=True)
                view[index].value = sumw
                view[index].variance = sumw2

        return out

    def to_hist(self):
        """Convert this coffea.hist histogram to a hist object"""
        import hist

        return hist.Hist(self.to_boost())



"""Histogramming tools

`coffea.hist` is a histogram filling, transformation, and plotting sub-package, utilizing
numpy arrays for storage and matplotlib plotting routines for visualization.

Features found in this package are similar to those found in
packages such as `histbook <https://github.com/scikit-hep/histbook>`__ (deprecated),
`boost-histogram <https://github.com/scikit-hep/boost-histogram>`__ (in development),
`physt <https://github.com/scikit-hep/boost-histogram>`__, and built-in numpy
`histogram <https://docs.scipy.org/doc/numpy/reference/generated/numpy.histogram.html>`__ utilities.

"""
from .hist_tools import (
    Hist,
    Bin,
    Interval,
    Cat,
    StringBin,
)
from .plot import (
    poisson_interval,
    clopper_pearson_interval,
    normal_interval,
    plot1d,
    plotratio,
    plot2d,
    plotgrid,
)
from .export import export1d

__all__ = [
    "Hist",
    "Bin",
    "Interval",
    "Cat",
    "StringBin",
    "poisson_interval",
    "clopper_pearson_interval",
    "normal_interval",
    "plot1d",
    "plotratio",
    "plot2d",
    "plotgrid",
    "export1d",
]



import awkward
from copy import copy
from coffea.nanoevents.methods.base import behavior, Systematic


@awkward.mixin_class(behavior)
class UpDownSystematic(Systematic):
    """An example instance of a simple systematic with only up/down variations."""

    _udmap = {"up": 0, "down": 1}

    def _build_variations(self, name, what, varying_function, *args, **kwargs):
        whatarray = (
            self[what] if what != "weight" else self["__systematics__", "__ones__"]
        )

        self["__systematics__", f"__{name}__"] = awkward.virtual(
            varying_function,
            args=(whatarray, *args),
            kwargs=kwargs,
            length=len(whatarray),
        )

    def describe_variations(self):
        """Show the map of variation names to indices."""
        return list(self._udmap.keys())

    def get_variation(self, name, what, astype, updown):
        """Calculate and up or down variation."""
        fields = awkward.fields(self)
        fields.remove("__systematics__")

        varied = self["__systematics__", f"__{name}__", :, self._udmap[updown]]

        params = copy(self.layout.parameters)
        params["variation"] = f"{name}-{what}-{updown}"

        out = {field: self[field] for field in fields}
        if what == "weight":
            out[f"weight_{name}"] = varied
        else:
            out[what] = varied

        return awkward.zip(
            out,
            depth_limit=1,
            parameters=params,
            behavior=self.behavior,
            with_name=astype,
        )

    def up(self, name, what, astype):
        """Return the "up" variation of this observable."""
        return awkward.virtual(
            self.get_variation,
            args=(name, what, astype, "up"),
            length=len(self),
            parameters=self[what].layout.parameters if what != "weight" else None,
        )

    def down(self, name, what, astype):
        """Return the "down" variation of this observable."""
        return awkward.virtual(
            self.get_variation,
            args=(name, what, astype, "down"),
            length=len(self),
            parameters=self[what].layout.parameters if what != "weight" else None,
        )


behavior[("__typestr__", "UpDownSystematic")] = "UpDownSystematic"
